/*
 * Created on Dec 1, 2004
 */
package weka.classifiers.ucd.biased;

import java.io.Serializable;
import java.text.DecimalFormat;
import java.util.Random;

import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.classifiers.evaluation.EvaluationUtils;
import weka.classifiers.ucd.eval.AbstractEvaluatingClassifier;
import weka.core.*;
import weka.core.AttributeStats;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.UnsupportedClassTypeException;

/**
 * @author andreas
 */
public class BiasSelector
extends AbstractEvaluatingClassifier
implements BiasedClassifier, Serializable {

    /**
     * In plot mode, normal search heuristics is suspended.
     * Instead, bias is raised stepwise from 0 to 1 and IR statistics
     * are gathered for each bias. This additional information is
     * output in the toString method.
     */
    protected boolean plotMode = false;
    
    protected StringBuffer gnuplot;
    
    /**
     * Initial bias setting
     */
    protected double startBias = 0.9;
    
    /**
     * Current bias setting
     */
    protected double bias = 0.9;
    
    /**
     * Previous bias
     */
    protected double previousBias = 0.9;
    
    /* The maximisation modes */
    public static final int MAXIMIZE_PRECISION = 0;
    public static final int MAXIMIZE_RECALL = 1;
    public static final int MAXIMIZE_F1 = 2;
    public static final Tag[] TAGS_MAXIMIZE = {
            new Tag(MAXIMIZE_PRECISION, "Maximize Precision"),
            new Tag(MAXIMIZE_RECALL, "Maximize Recall"),
            new Tag(MAXIMIZE_F1, "Maximize F1")
    };
    
    /**
     * Which measure to maximize for classIndex
     */
    protected int maximize = MAXIMIZE_PRECISION;
    
    /* The constrain modes */
    public static final int CONSTRAIN_PRECISION = 0;
    public static final int CONSTRAIN_RECALL = 1;
    public static final int CONSTRAIN_F1 = 2;
    public static final int CONSTRAIN_ACCURACY = 3;
    // public static final int CONSTRAIN_MAVG_F1 = 4; // future option
    public static final Tag[] TAGS_CONSTRAIN = {
            new Tag(CONSTRAIN_PRECISION, "Constrain precision"),
            new Tag(CONSTRAIN_RECALL, "Constrain recall"),
            new Tag(CONSTRAIN_F1, "Constrain F1"),
            new Tag(CONSTRAIN_ACCURACY, "Constrain overall accuracy")
    };
    
    /**
     * Which measure to put a minimum constraint on
     */
    protected int constrain = CONSTRAIN_RECALL;
    
    /**
     * The minimum value for the constraint measure
     */
    protected double minimum = 0.05;
    
    /* (non-Javadoc)
     * @see weka.classifiers.ucd.biased.BiasedClassifier#getPreferredClassIndex()
     */
    public int getPreferredClassIndex() {
        return classIndex;
    }

    /* (non-Javadoc)
     * @see weka.classifiers.ucd.biased.BiasedClassifier#setPreferredClassIndex(int)
     */
    public void setPreferredClassIndex(int i) {
        classIndex = i;
    }

    /* (non-Javadoc)
     * @see weka.classifiers.ucd.biased.BiasedClassifier#getBiasStrength()
     */
    public double getBiasStrength() {
        return bias;
    }

    /* (non-Javadoc)
     * @see weka.classifiers.ucd.biased.BiasedClassifier#setBiasStrength(double)
     */
    public void setBiasStrength(double d) {
        // not useful here
    }
    
    /* (non-Javadoc)
     * @see weka.classifiers.Classifier#buildClassifier(weka.core.Instances)
     */
    public void buildClassifier(Instances instances) throws Exception {
        
        if (!(classifier instanceof BiasedClassifier)) {
            throw new Exception("Classifier is not an instance of BiasedClassifier");
        }
        
        int nClasses = instances.numClasses();
        
//        if (nClasses > 2) {
//            throw new UnsupportedClassTypeException("Only works for two-class datasets!");
//        }
        
        if (!instances.classAttribute().isNominal()) {
            throw new UnsupportedClassTypeException("Class attribute must be nominal!");
        }
        
        AttributeStats stats = instances.attributeStats(instances.classIndex());

        // If data contains only one instance of positive data
        // optimize on training data

        if (stats.distinctCount < 2) {
            System.err.println("Couldn't find examples of both classes. No adjustment.");
            classifier.buildClassifier(instances);
            return;
        }
        
        // Initialize everything
        
        rnd = new Random(seed);
        if (plotMode) {
            bias = 0;
        } else {
            bias = startBias;
        }
        
        double step = 1.0 / ((double) maxRounds - 1.0);
        double constrainedBestBias = 0;
        
        gnuplot = new StringBuffer();
        
        StringBuffer precision[] = new StringBuffer[nClasses];
        StringBuffer recall[] = new StringBuffer[nClasses];
        StringBuffer f1[] = new StringBuffer[nClasses];
        
        DecimalFormat dfBias = new DecimalFormat("0.00");
        DecimalFormat dfValue = new DecimalFormat("0.######");
        
        for (int n = 0; n < nClasses; n++) {
            precision[n] = new StringBuffer();
            precision[n].append("# Bias vs. Precision for class " + n + "\n");
            recall[n] = new StringBuffer();
            recall[n].append("# Bias vs. Recall for class " + n + "\n");
            f1[n] = new StringBuffer();
            f1[n].append("# Bias vs. F-Measure for class " + n + "\n");
        }
        
        StringBuffer acc = new StringBuffer();
        acc.append("# Bias vs. Accuracy\n");

        // loop: iteration rounds
        
        double bestF1 = 0;
        double bestPrec = 0;
        
        for (int n = 0; n < maxRounds; n++) {
            
            if (plotMode) {
                // In plot mode, raise bias stepwise
                bias = step * n;
            }
            
            ((BiasedClassifier) classifier).setBiasStrength(bias);
            ((BiasedClassifier) classifier).setPreferredClassIndex(classIndex);

            Evaluation ev = getEvaluation(instances);
            
            for (int m = 0; m < nClasses; m++) {
                precision[m].append(dfBias.format(bias) + " " + dfValue.format(ev.precision(m)) + "\n");
                recall[m].append(dfBias.format(bias) + " " + dfValue.format(ev.recall(m)) + "\n");
                f1[m].append(dfBias.format(bias) + " " + dfValue.format(ev.fMeasure(m)) + "\n");
            }
            acc.append(dfBias.format(bias) + " " + dfValue.format(ev.pctCorrect() / 100) + "\n");
            
            boolean constraint = true;
            
            switch(constrain) {
            case CONSTRAIN_PRECISION:
                if (ev.precision(classIndex) < minimum) constraint = false;
                break;
            case CONSTRAIN_RECALL:
                if (ev.recall(classIndex) < minimum) constraint = false;
                break;
            case CONSTRAIN_F1:
                if (ev.fMeasure(classIndex) < minimum) constraint = false;
                break;
            case CONSTRAIN_ACCURACY:
                if (ev.pctCorrect() / 100 < minimum) constraint = false;
                break;
            }
            
            switch(maximize) {
            case MAXIMIZE_PRECISION:
                // Heuristic for maximising precision:
                // Move bias towards 1
                if (constraint) {
                    if (ev.precision(classIndex) > bestPrec) {
                        bestPrec = ev.precision(classIndex);
                        constrainedBestBias = bias;
                    }
	                bias = 1 - (( 1 - bias) / 2);
                } else {
                    bias = constrainedBestBias + ((bias - constrainedBestBias) / 2);
                }
                break;
            case MAXIMIZE_F1:
                if (constraint) {
                    if (ev.fMeasure(classIndex) > bestF1) {
                        bestF1 = ev.fMeasure(classIndex);
                        constrainedBestBias = bias;
                    }
                }
                // no heuristic, use only with plot mode
                break;
            default:
                throw new Exception("Sorry, I don't have a heuristic for that yet.");
            }
        }
        
        // use the last bias that was known to be safe
        
        bias = constrainedBestBias;
        ((BiasedClassifier) classifier).setBiasStrength(bias);
        ((BiasedClassifier) classifier).setPreferredClassIndex(classIndex);
        classifier.buildClassifier(instances);
        
        // compile gnuplot output
        
        for (int n = 0; n < nClasses; n++) {
            gnuplot.append(precision[n].toString());
            gnuplot.append("\n");
            gnuplot.append(recall[n].toString());
            gnuplot.append("\n");
            gnuplot.append(f1[n].toString());
            gnuplot.append("\n");
        }
        gnuplot.append(acc.toString());
        gnuplot.append("\n");
    }
    
    public double [] distributionForInstance(Instance instance) throws Exception {
        return classifier.distributionForInstance(instance);
    }

    public String toString() {
        try {
            StringBuffer output = new StringBuffer();
            output.append(this.getClass().getName());
            output.append("\nClass index: " + classIndex + " Bias: " + bias + "\n");
            output.append("\nInner classifier: " + classifier.getClass().getName() + "\n");
            output.append("\n");
            output.append("--- IR statistics vs. bias:\n\n");
            output.append(gnuplot.toString());
            output.append("\n");
            output.append(classifier.toString());
            return output.toString();
        } catch (RuntimeException e) {
            return "No model built yet?";
        }
    }
    /**
     * @return Returns the constrain.
     */
    public SelectedTag getConstrain() {
        return new SelectedTag(constrain, TAGS_CONSTRAIN);
    }
    /**
     * @param constrain The constrain to set.
     */
    public void setConstrain(SelectedTag constrain) {
        if (constrain.getTags() == TAGS_CONSTRAIN) {
            this.constrain = constrain.getSelectedTag().getID();
        }
    }
    /**
     * @param evaluationMode The evaluationMode to set.
     */
    public void setEvaluationMode(SelectedTag newMethod) {
        if (newMethod.getTags() == TAGS_EVAL) {
            evaluationMode = newMethod.getSelectedTag().getID();
        }
    }
    
    /**
     * Gets the evaluation mode used. Will be one of
     * EVAL_TRAINING, EVAL_TUNED_SPLIT, or EVAL_CROSS_VALIDATION
     *
     * @return the evaluation mode.
     */
    public SelectedTag getEvaluationMode() {

      return new SelectedTag(evaluationMode, TAGS_EVAL);
    }

    /**
     * @return Returns the maximize.
     */
    public SelectedTag getMaximize() {
        return new SelectedTag(maximize, TAGS_MAXIMIZE);
    }
    /**
     * @param maximize The maximize to set.
     */
    public void setMaximize(SelectedTag maximize) {
        if (maximize.getTags() == TAGS_MAXIMIZE) {
            this.maximize = maximize.getSelectedTag().getID();
        }
        if (this.maximize == MAXIMIZE_F1) plotMode = true;
    }
    /**
     * @return Returns the minimum.
     */
    public double getMinimum() {
        return minimum;
    }
    /**
     * @param minimum The minimum to set.
     */
    public void setMinimum(double minimum) {
        this.minimum = minimum;
    }
    /**
     * @return Returns the startBias.
     */
    public double getInitialBias() {
        return startBias;
    }

    /**
     * @param startBias The startBias to set.
     */
    public void setInitialBias(double startBias) {
        this.startBias = startBias;
    }
    /**
     * @return Returns the plotMode.
     */
    public boolean isPlotMode() {
        return plotMode;
    }
    /**
     * @param plotMode The plotMode to set.
     */
    public void setPlotMode(boolean plotMode) {
        this.plotMode = plotMode;
    }

    /**
     * @return Returns the classifier.
     */
    public Classifier getClassifier() {
        return classifier;
    }

    /**
     * @param classifier The classifier to set.
     */
    public void setClassifier(Classifier classifier) {
        this.classifier = classifier;
    }
}
