/*
 * Created on Nov 25, 2004
 */
package weka.classifiers.ucd.biased;

import java.io.Serializable;
import java.util.Enumeration;
import java.util.Vector;

import weka.classifiers.Classifier;
import weka.classifiers.functions.SMO;
import weka.core.*;

/**
 * Abstract base class for classifiers that can be biased for
 * increased precision for one particular class.
 * @author andreas
 */
public abstract class BiasedClassifierImpl
	extends Classifier
	implements OptionHandler, Serializable, BiasedClassifier {

	/**
	 * The base classifier
	 */
    protected Classifier classifier = new SMO();
	
    /**
     * The interpretation of the concrete bias value depends on the
     * implementation of the buildClassifier method.
     */
	protected double bias = 1.0;
	
	/**
	 * Implementations of a BiasedClassifier are supposed to be biased
	 * towards high PRECISION for this class index.
	 */
	protected int classIndex = 0;

	/* (non-Javadoc)
	 * @see weka.classifiers.Classifier#buildClassifier(weka.core.Instances)
	 */
	public abstract void buildClassifier(Instances originalDataset) throws Exception;
	
	/* (non-Javadoc)
	 * @see weka.classifiers.ucd.ensemble.Biased#getBias()
	 */
	public double getBiasStrength() {
		return bias;
	}

	/* (non-Javadoc)
	 * @see weka.classifiers.ucd.ensemble.Biased#setBias(double)
	 */
	public void setBiasStrength(double d) {
		bias = d;
	}

	/**
	 * @return
	 */
	public Classifier getClassifier() {
		return classifier;
	}

	/**
	 * @param classifier
	 */
	public void setClassifier(Classifier _classifier) {
		classifier = _classifier;
	}

	/* (non-Javadoc)
	 * @see weka.core.OptionHandler#getOptions()
	 */
	public String[] getOptions() {
		String [] classifierOptions = new String [0];
		if (classifier != null) {
		  classifierOptions = ((OptionHandler) classifier).getOptions();
		}
		String [] options = new String [classifierOptions.length + 10];
		int c = 0;
		
		options[c++] = "-W"; options[c++] = classifier.getClass().getName();
		options[c++] = "-B"; options[c++] = "" + bias;
		options[c++] = "-C"; options[c++] = "" + classIndex;

		System.arraycopy(classifierOptions, 0, options, c, 
				 classifierOptions.length);
		c += classifierOptions.length;
		while (c < options.length) {
		  options[c++] = "";
		}

		return options;
	}

	/* (non-Javadoc)
	 * @see weka.core.OptionHandler#listOptions()
	 */
	public Enumeration listOptions() {
		Vector options = new Vector();
		options.add(
				new Option("\tSpecify the full class name of a base classifier " +
						"(default NaiveBayes).", "W", 1, "-W <classname>"));
		options.add(
				new Option("\tBias towards this class index",
						"C", 1, "-Q <int>"));
		options.add(
				new Option("\tDegree of bias (interpretation depends on implementation)",
						"B", 1, "-B <double>"));
		return options.elements();
	}

	/* (non-Javadoc)
	 * @see weka.core.OptionHandler#setOptions(java.lang.String[])
	 */
	public void setOptions(String[] options) throws Exception {
		String classifierNameOption = Utils.getOption('W', options);
		
		String biasOption = Utils.getOption('B', options);
		if (biasOption.length() != 0 ) {
			bias = Double.parseDouble(biasOption);
		}

		String classIndexOption = Utils.getOption('C', options);
		if (classIndexOption.length() != 0) {
			classIndex = Integer.parseInt(classIndexOption);
		}

		String[] classifierOptions = Utils.partitionOptions(options);
		if (classifierNameOption.length() != 0) {
			classifier = Classifier.forName(classifierNameOption, classifierOptions);
		}
	}

	/**
	 * @return
	 */
	public int getPreferredClassIndex() {
		return classIndex;
	}

	/**
	 * @param i
	 */
	public void setPreferredClassIndex(int i) {
		classIndex = i;
	}

	/* (non-Javadoc)
	 * @see weka.classifiers.Classifier#distributionForInstance(weka.core.Instance)
	 */
	public double[] distributionForInstance(Instance instance)
		throws Exception {
		return classifier.distributionForInstance(instance);
	}

    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString() {
        StringBuffer output = new StringBuffer();
        output.append(this.getClass().getName());
        output.append("\nClass index: " + classIndex + " Bias: " + bias + "\n");
        output.append("\nInner classifier: " + classifier.getClass().getName() + "\n");
        output.append("\n");
        output.append(classifier.toString());
        return output.toString();
    }
}
