/*
 * Created on 18.12.2004
 */
package weka.classifiers.ucd.biased;

import java.io.Serializable;

import weka.classifiers.Classifier;
import weka.classifiers.evaluation.ConfusionMatrix;
import weka.classifiers.evaluation.TwoClassStats;
import weka.classifiers.functions.SMO;
import weka.classifiers.ucd.eval.RelabelUtils;
import weka.core.AttributeStats;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.MakeIndicator;

/**
 * Algorithm that tries to cover negative examples (with classIndex !=
 * preferredClassIndex). Uses multiple instances of a base classifier
 * to perform covering. Stops if the fraction of covered negatives is
 * >= biasStrength or until a maximum number of iterations has been reached.
 * @author andreas
 */
public class CoverNegatives extends Classifier
implements BiasedClassifier, Serializable {
	
	/**
	 * Fraction of negatives that have to be covered for the covering to stop
	 */
	protected double biasStrength = 0.9;
	
	/**
	 * Fraction of negatives that are covered
	 */
	protected double fCovered = 0;
	
	/**
	 * Number of negative instances that are covered
	 */
	protected int nCovered = 0;
	
	/**
	 * Total number of negative instances
	 */
	protected int nNegatives = 0;
	
	/**
	 * Class index of positives
	 */
	protected int preferredClassIndex = 0;
	
	/**
	 * Maximum number of iterations for covering
	 */
	protected int maxIterations = 10;
	
	/**
	 * Actual number of iterations performed
	 */
	protected int iteration = 0;
	
	/**
	 * Base classifier
	 */
	protected Classifier baseClassifier = new SMO();
	
	/**
	 * Classifiers for covering
	 */
	protected Classifier[] classifiers;
	
	/**
	 * Output for toString
	 */
	protected StringBuffer output;
	
	/**
	 * Class index that is returned as negative class
	 * See comment in classifyInstance for details
	 */
	protected int negativeClassIndex;

	/* (non-Javadoc)
	 * @see weka.classifiers.Classifier#buildClassifier(weka.core.Instances)
	 */
	public void buildClassifier(Instances originalTrainingSet)
	throws Exception {
		
		// Relabel to binary dataset
		// All instances of the preferred class are relabeled as 1,
		// all others as 0.
		
		MakeIndicator classFilter = new MakeIndicator();
		classFilter.setValueIndex(preferredClassIndex);
		classFilter.setNumeric(false);
		classFilter.setInputFormat(originalTrainingSet);
		Instances trainingSet =
			Filter.useFilter(originalTrainingSet, classFilter);
		
		// Initialize matrix for relabeling the instances for each iteration
		// of covering. In each iteration, instances classified as negatives
		// are dropped for the next iteration. Instances that were classified
		// as positives are kept and covering continues.
		
		double[][] relabelMatrix = new double[2][2];
		relabelMatrix[0][0] = -1;
		relabelMatrix[0][1] = -1;
		relabelMatrix[1][0] = 0;
		relabelMatrix[1][1] = 1;
		
		// Initialize classifiers for covering
		
		classifiers = Classifier.makeCopies(baseClassifier, maxIterations);
		
		// Intialize fields
		
		AttributeStats stats = trainingSet.attributeStats(trainingSet.classIndex());
		nNegatives = stats.nominalCounts[0];
		fCovered = 0;
		nCovered = 0;
		iteration = 0;
		output = new StringBuffer();
		negativeClassIndex = 0;
		int trueNegatives = 0;
		if (preferredClassIndex == 0) negativeClassIndex = 1;
		
		// Initialize class labels for confusion matrix
		
		String[] classLabels = {"negative", "positive"};

		// Iterate covering until we have either covered enough negative
		// instances or the maximum number of iterations has been reached.
		
		do {
			// Train, classify and relabel
			
			Instances relabeledInstances = new Instances(trainingSet, 0);
			RelabelUtils ru = new RelabelUtils(relabeledInstances, relabelMatrix);
			FastVector predictions =
				ru.getTrainTestPredictions(
						classifiers[iteration], trainingSet, trainingSet);
			
			// Evaluate
			
			ConfusionMatrix cm = new ConfusionMatrix(classLabels);
			cm.addPredictions(predictions);
			TwoClassStats tcs = cm.getTwoClassStats(1);
			
			// See how many new negatives were covered
			
			trueNegatives = (int) tcs.getTrueNegative();
			
			if (trueNegatives == 0) {
				// No new negatives covered, break.
				output.append("No new instances covered in iteration " +
						iteration + "\n");
				break;
			}
			
			nCovered += trueNegatives;
			fCovered = (double) nCovered / (double) nNegatives;
			
			// Optional debug output
			
			output.append("Iteration " + iteration +
						", covered " + nCovered + " of " + nNegatives +
						", positives left: " + tcs.getTruePositive() + "\n");
			
			// Copy relabeled instances for next round
			
			trainingSet = relabeledInstances;
			
			// advance to next round
			
			iteration++;
			
			// continue if not enough instances covered and
			// max number of iterations not reached and
			// any new instances covered in last round
			
		} while (fCovered < biasStrength && iteration < maxIterations);
	}

	/* (non-Javadoc)
	 * @see weka.classifiers.Classifier#classifyInstance(weka.core.Instance)
	 */
	public double classifyInstance(Instance instance) throws Exception {
		
		// Loop all initialized classifiers, but return if one of the
		// classifiers covers the instance as negative
		
		for (int n = 0; n < iteration; n++) {
			if (classifiers[n].classifyInstance(instance) == 0) {
				
				// Classifier predicted 0, i.e. this instance is covered
				// as a negative. Because we've trained on a binary dataset
				// only, however, we cannot predict the true class of the
				// instance! Return negativeClassIndex instead.
				
				return negativeClassIndex;
			}
		}
		
		// The instance was not covered as a negative.
		
		return preferredClassIndex;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "CoverNegatives\n\n" + output.toString();
	}
	
	/**
	 * @return Returns the biasStrength.
	 */
	public double getBiasStrength() {
		return biasStrength;
	}
	
	/**
	 * @param biasStrength The biasStrength to set.
	 */
	public void setBiasStrength(double biasStrength) {
		this.biasStrength = biasStrength;
	}
	
	/**
	 * @return Returns the maxIterations.
	 */
	public int getMaxIterations() {
		return maxIterations;
	}
	
	/**
	 * @param maxIterations The maxIterations to set.
	 */
	public void setMaxIterations(int maxIterations) {
		this.maxIterations = maxIterations;
	}
	
	/**
	 * @return Returns the preferredClassIndex.
	 */
	public int getPreferredClassIndex() {
		return preferredClassIndex;
	}
	
	/**
	 * @param preferredClassIndex The preferredClassIndex to set.
	 */
	public void setPreferredClassIndex(int preferredClassIndex) {
		this.preferredClassIndex = preferredClassIndex;
	}
	
	/**
	 * @return Returns the baseClassifier.
	 */
	public Classifier getBaseClassifier() {
		return baseClassifier;
	}
	
	/**
	 * @param baseClassifier The baseClassifier to set.
	 */
	public void setBaseClassifier(Classifier baseClassifier) {
		this.baseClassifier = baseClassifier;
	}
}
