/*
 * Created on Dec 2, 2004
 */
package weka.classifiers.ucd.biased;

import java.io.Serializable;
import java.util.Enumeration;
import java.util.Random;

import weka.classifiers.Classifier;
import weka.classifiers.evaluation.ConfusionMatrix;
import weka.classifiers.evaluation.TwoClassStats;
import weka.classifiers.functions.SMO;
import weka.classifiers.ucd.eval.AbstractEvaluatingClassifier;
import weka.core.*;

/**
 * Meta-classifier that generates a biased classifier based on the
 * misclassifications of an unbiased classifier.
 * 
 * Misclassified instances of the class that we are biasing towards high
 * precision are re-labeled as instances of the other class, and the biased
 * classifier is trained on the re-labeled set.
 * 
 * This phase can be repeated until, for example, precision for the biased
 * class reaches 1 or some other termination criterion (like a maximum number
 * of rounds) is met. 
 * 
 * @author andreas
 */
public class ErrorBiasedClassifier extends AbstractEvaluatingClassifier
        implements BiasedClassifier, Serializable {
    
    /**
     * Last evaluated precision on the preferred class 
     */
    protected double precision;
    
    /**
     * Last evaluated recall on the preferred class
     */
    protected double recall;
    
    /**
     * The minimum number of instances in the preferred class
     * This is an absolute number.
     */
    protected int minInstances = 3;
    
    /**
     * The minimum recall for the preferred class
     */
    protected double minRecall = 0.05;
    
    /**
     * Terminate loop, if precision > this value
     */
    protected double goalPrecision = 0.999;
    
    /**
     * Minimum precision: If best precision < this value,
     * always predict negative class
     */
    protected double minPrecision = 0.85;
    
    /**
     * Output for toString, gives information about the building phase
     */
    protected StringBuffer output;
    
    protected int nClasses;
    
    public ErrorBiasedClassifier() {
        // different default classifier than the abstract super class
        classifier = new SMO();
    }
    
    public String toString() {
        try {
            return new String("ErrorBiasClassifier\n\n" +
                    output.toString() + "\n\nBase classifier:" + 
                    classifier.toString());
        } catch (RuntimeException e) {
            return "No model built yet.";
        }
    }
    
    /* (non-Javadoc)
     * @see weka.classifiers.Classifier#buildClassifier(weka.core.Instances)
     */
    public void buildClassifier(Instances instances) throws Exception {
        
        output = new StringBuffer();
        
        if (!instances.classAttribute().isNominal()) {
            throw new UnsupportedClassTypeException("Class attribute must be nominal!");
        }
        
        AttributeStats stats = instances.attributeStats(instances.classIndex());

        // If data contains only one instance of positive data
        // optimize on training data

        if (stats.distinctCount < 2) {
            System.err.println("Couldn't find examples of both classes. No adjustment.");
            classifier.buildClassifier(instances);
            return;
        }
       
        // Initialize everything
    
        nClasses = instances.numClasses();
        rnd = new Random(seed);
        Instances lastInstanceSet = new Instances(instances);
        double bestPrecision = 0;
        Instances bestInstanceSet = lastInstanceSet;
        String[] classLabels = new String[instances.numClasses()];
        Enumeration e = instances.classAttribute().enumerateValues();
        for (int n = 0; n < classLabels.length && e.hasMoreElements(); n++) {
            classLabels[n] = (String) e.nextElement();
        }
        
        // loop: iteration rounds
        
        for (int n = 0; n < maxRounds; n++) {
            FastVector predictions = getPredictions(lastInstanceSet, true);
            ConfusionMatrix c = new ConfusionMatrix(classLabels);
            c.addPredictions(predictions);
            TwoClassStats t = c.getTwoClassStats(classIndex);
            double truePositive = t.getTruePositive();
            double currentRecall = t.getRecall();
            double currentPrecision = t.getPrecision();
            output.append(
                        "Round " + n +
                        " TP: " + truePositive +
                        " FP: " + t.getFalsePositive() + 
                        " P: " + currentPrecision +
                        " R: " + currentRecall + "\n");
            if (truePositive < minInstances) {
                output.append("TP < minInstances, break.\n");
                break;
            }
            if (currentRecall < minRecall) {
                output.append("R < minRecall, break.\n");
                break;
            }
//            if (currentPrecision <= precision) {
//                output.append("P <= precision, break.\n");
//                break;
//            }

            // constraints are met
            
            precision = currentPrecision;
            recall = currentRecall;
            lastInstanceSet = relabeledInstances;
            
            if (currentPrecision > bestPrecision) {
                bestPrecision = currentPrecision;
                bestInstanceSet = relabeledInstances;
            }
            
            if (precision > goalPrecision) {
                output.append("P > goalPrecision, break.\n");
                break;
            }
        }
        
        // build classifier on last instance set known as satisfying
        // the constraints
        
        precision = bestPrecision;
        if (precision > minPrecision)
        	classifier.buildClassifier(bestInstanceSet);
       
   }

    /* (non-Javadoc)
     * @see weka.classifiers.ucd.biased.BiasedClassifier#getPreferredClassIndex()
     */
    public int getPreferredClassIndex() {
        return classIndex;
    }

    /* (non-Javadoc)
     * @see weka.classifiers.ucd.biased.BiasedClassifier#setPreferredClassIndex(int)
     */
    public void setPreferredClassIndex(int i) {
        classIndex = i;
    }

    /* (non-Javadoc)
     * @see weka.classifiers.ucd.biased.BiasedClassifier#getBiasStrength()
     */
    public double getBiasStrength() {
        // We return the last precision of on the preferred class as the
        // current bias strength. Note that it does *not* make sense to *set*
        // the bias strength. This class is about figuring out a biased
        // classifier without specifying a bias strength in advance.
        return precision;
    }

    /* (non-Javadoc)
     * @see weka.classifiers.ucd.biased.BiasedClassifier#setBiasStrength(double)
     */
    public void setBiasStrength(double d) {
        // Ignore:
        // Bias strength is meaningless here, because we are figuring out
        // a biased classifier without wanting to specify a bias strength
        // in advance.
    }
    
    public double[] distributionForInstance(Instance i) throws Exception {
    	if (precision < minPrecision) {
    		// if precision is too low, never predict preferred class
    		double[] distr = new double[nClasses];
    		for (int n = 0; n < nClasses; n++) {
    			distr[n] = 1.0 / ((double) nClasses - 1.0);
    		}
			distr[classIndex] = 0;
			return distr;
    	}
        return classifier.distributionForInstance(i); 
 
    }

	public double getMinPrecision() {
		return minPrecision;
	}
	public void setMinPrecision(double minPrecision) {
		this.minPrecision = minPrecision;
	}
	public double getGoalPrecision() {
		return goalPrecision;
	}
	public void setGoalPrecision(double goalPrecision) {
		this.goalPrecision = goalPrecision;
	}
	public int getMinInstances() {
		return minInstances;
	}
	public void setMinInstances(int minInstances) {
		this.minInstances = minInstances;
	}
	public double getMinRecall() {
		return minRecall;
	}
	public void setMinRecall(double minRecall) {
		this.minRecall = minRecall;
	}

    /**
     * @return Returns the classifier.
     */
    public Classifier getClassifier() {
        return classifier;
    }

    /**
     * @param classifier The classifier to set.
     */
    public void setClassifier(Classifier classifier) {
        this.classifier = classifier;
    }
}
