/*
 * Created on Dec 7, 2004
 */
package weka.classifiers.ucd.biased;

import java.util.*;

import weka.classifiers.Classifier;
import weka.classifiers.functions.SMO;
import weka.core.*;

/**
 * @author andreas
 */
public class RandomBiasedEnsemble extends Classifier implements BiasedClassifier, Randomizable {

    /**
     * Comparable class for sorting instances by distance
     * @author andreas
     */
    class SortedInstance implements Comparable {
        
        double distance;
        Instance instance;

        /**
         * Compute the euclidian distance between to instances
         * Borrowed from IB1, but does not normalise.
         * @param i
         * @param j
         * @return
         */
        private double distance(Instances dataset, Instance first, Instance second) {
            
            double diff, distance = 0;
            
            for(int i = 0; i < dataset.numAttributes(); i++) { 
                if (i == dataset.classIndex()) {
                    continue;
                }
                if (dataset.attribute(i).isNominal()) {
                    
                    // If attribute is nominal
                    if (first.isMissing(i) || second.isMissing(i) ||
                            ((int)first.value(i) != (int)second.value(i))) {
                        distance += 1;
                    }
                } else {
                    
                    // If attribute is numeric
                    if (first.isMissing(i) || second.isMissing(i)){
                        if (first.isMissing(i) && second.isMissing(i)) {
                            diff = 1;
                        } else {
                            if (second.isMissing(i)) {
                                // diff = norm(first.value(i), i);
                                diff = first.value(i);
                            } else {
                                // diff = norm(second.value(i), i);
                                diff = second.value(i);
                            }
                            if (diff < 0.5) {
                                diff = 1.0 - diff;
                            }
                        }
                    } else {
                        // diff = norm(first.value(i), i) - norm(second.value(i), i);
                        diff = first.value(i) - second.value(i);
                    }
                    distance += diff * diff;
                }
            }
            
            return distance;
          }
        
        SortedInstance(Instances dataset, Instance distanceTo, Instance i) {
            instance = i;
            distance = distance(dataset, distanceTo, i);
        }

        /* (non-Javadoc)
         * @see java.lang.Comparable#compareTo(java.lang.Object)
         */
        public int compareTo(Object o) {
            if (o instanceof SortedInstance) {
                SortedInstance s = (SortedInstance) o;
                if (s.getDistance() < distance) return 	1;
                if (s.getDistance() > distance) return 	-1;
                return 0;                
            }
            return 0;
        }
        
        public double getDistance() {
            return distance;
        }
        
        public Instance getInstance() {
            return instance;
        }
    }

    protected Classifier classifier = new SMO();
    
    protected int preferredClass = 0;
    
    protected int ensembleSize = 10;
    
    protected int sampleSizePositive = 20;
    
    protected int sampleSizeNegative = 40;
    
    protected Classifier[] ensemble;
    
    /**
     * Seed for random number generator
     */
    protected int seed = 1;
    
    /**
     * Becomes true, if something happens during training and the classifier
     * can not be build in the ensemble way. Example: If there are too few
     * positive instances. The base classifier will then be trained the
     * default way and the classifyInstance method will return the result
     * from the base classifier directly.
     */
    protected boolean fallback = false;

    /* (non-Javadoc)
     * @see weka.classifiers.Classifier#buildClassifier(weka.core.Instances)
     */
    public void buildClassifier(Instances instances) throws Exception {
        
        // Check preconditions
        
        if (!instances.classAttribute().isNominal()) {
            throw new UnsupportedClassTypeException("Class attribute must be nominal!");
        }
        
        AttributeStats stats = instances.attributeStats(instances.classIndex());
        
        // Revert to "failsafe" mode if too few instances
        
        if (stats.nominalCounts[preferredClass] < sampleSizePositive) {
            // too few positive instances
            fallback = true;
            classifier.buildClassifier(instances);
        }
        
        ensemble = Classifier.makeCopies(classifier, ensembleSize); // Initialize ensemble
        
        // Sort instances
        
        Instances positives = new Instances(instances, 0);
        Instances negatives = new Instances(instances, 0);
        Enumeration ex = instances.enumerateInstances();
        while (ex.hasMoreElements()) {
            Instance i = (Instance) ex.nextElement();
            if (i.classValue() == preferredClass)
                positives.add(i);
            else
                negatives.add(i);
        }
        
        // Randomize instances
        
        // Random rnd = new Random(seed);
        Random rnd = new Random();
        positives.randomize(rnd);
        negatives.randomize(rnd);
        
        // Get first positive and calc distance to others
        
        SortedSet closestInstances = new TreeSet();
        Enumeration ep = positives.enumerateInstances();
        Instance reference = null; // first instance is "seed" for the set
        if (ep.hasMoreElements()) reference = (Instance) ep.nextElement();
        while (ep.hasMoreElements()) {
            closestInstances.add(
                    new SortedInstance(instances, reference,
                            (Instance) ep.nextElement()));
        }
        
        // Train ensemble
        
        Enumeration en = negatives.enumerateInstances();
        for (int n = 0; n < ensembleSize; n++) {
            Instances dataset = new Instances(instances, 0);
            dataset.add(reference);
            
            // add closest positives
            
            Iterator it = closestInstances.iterator();
            int m = sampleSizePositive;
            while(it.hasNext() && --m > 0) {
                dataset.add(((SortedInstance) it.next()).instance);
            }
            
            // add random negatives
            
            for (int j = 0; j < sampleSizeNegative; j++) {
                if (!en.hasMoreElements()) {
                    en = negatives.enumerateInstances();
                    if (!en.hasMoreElements()) {
                        // no negatives!!
                        fallback = true;
                        classifier.buildClassifier(instances);
                    }
                }
                dataset.add((Instance) en.nextElement());
            }
            
            // train classifier
            
            ensemble[n].buildClassifier(dataset);
        }
    }
    
    public double classifyInstance(Instance instance) throws Exception {
        
        // failsafe mode, if ensemble could not be constructed for some reason
        
        if (fallback) return classifier.classifyInstance(instance);
        
        // return preferred class if and only if all ensemble members agree
        // if ensemble members disagree, return a random negative class
        
        // System.out.print("Individual predictions:");
        
        for (int n = 0; n < ensembleSize; n++) {
            double prediction = ensemble[n].classifyInstance(instance);
            // System.out.print(" " + prediction);
            if (prediction != preferredClass) {
                // System.out.println();
                return prediction;
            }
        }
        
        // System.out.println();
        return preferredClass;
        
    }
    
    /* (non-Javadoc)
     * @see weka.classifiers.ucd.biased.BiasedClassifier#getPreferredClassIndex()
     */
    public int getPreferredClassIndex() {
        return preferredClass;
    }

    /* (non-Javadoc)
     * @see weka.classifiers.ucd.biased.BiasedClassifier#setPreferredClassIndex(int)
     */
    public void setPreferredClassIndex(int i) {
        preferredClass = i;
    }

    /* (non-Javadoc)
     * @see weka.classifiers.ucd.biased.BiasedClassifier#getBiasStrength()
     */
    public double getBiasStrength() {
        return 0;
    }

    /* (non-Javadoc)
     * @see weka.classifiers.ucd.biased.BiasedClassifier#setBiasStrength(double)
     */
    public void setBiasStrength(double d) {
        // ignore
    }

    public Classifier getClassifier() {
        return classifier;
    }
    
    public void setClassifier(Classifier classifier) {
        this.classifier = classifier;
    }

    /* (non-Javadoc)
     * @see weka.core.Randomizable#setSeed(int)
     */
    public void setSeed(int _seed) {
        seed = _seed;
    }

    /* (non-Javadoc)
     * @see weka.core.Randomizable#getSeed()
     */
    public int getSeed() {
        return seed;
    }
    public int getEnsembleSize() {
        return ensembleSize;
    }
    public void setEnsembleSize(int ensembleSize) {
        this.ensembleSize = ensembleSize;
    }
    public int getSampleSizeNegative() {
        return sampleSizeNegative;
    }
    public void setSampleSizeNegative(int sampleSizeNegative) {
        this.sampleSizeNegative = sampleSizeNegative;
    }
    public int getSampleSizePositive() {
        return sampleSizePositive;
    }
    public void setSampleSizePositive(int sampleSizePositive) {
        this.sampleSizePositive = sampleSizePositive;
    }
}
