/*
 * Created on 02.10.2004
 */
package weka.classifiers.ucd.ensemble;

import java.io.*;
import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.Serializable;
import java.util.Enumeration;
import java.util.Vector;

import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.classifiers.functions.SMO;
import weka.classifiers.trees.J48;
import weka.classifiers.ucd.biased.BiasedClassifier;
import weka.classifiers.ucd.biased.UndersamplingBiasedClassifier;
import weka.core.*;
import weka.filters.unsupervised.attribute.MakeIndicator;

/**
 * @author andreas
 */
public class Triskel extends Classifier implements OptionHandler, Serializable {
    
    double bias0 = 0.9;
    double bias1 = 0.9;
    boolean separateBiases = false;
	
	/**
	 * some additional statistics
	 * element 0-(nClasses-1): individual classifiers
	 * element nClasses: arbiter
	 * element nClasses+1: arbiter (when actually used)
	 * element nClasses+2: -1 if disagree, = element 0 otherwise 
	 */
	double detailedResults[];

	/**
	 * Subclassifiers
	 */
	Classifier[] clc;
	
	/**
	 * Subclassifier
	 */
	Classifier[] arbiter;
	
	/**
	 * If rounds > 1, then the arbiter will be replaced with another
	 * Triskel with the same parameters as this one except
	 * that rounds will be decreased by 1.
	 */
	int rounds = 1;

	/**
	 * The number of instances the arbiter was trained on in the last round.
	 * If the arbiter in the next round is trained on the same number of
	 * instances (ie. the biased classifiers disagreed on every instance!)
	 * then further rounds are suspended.
	 */	
	int arbiterInstancesInLastRound = -1;

	/**
	 * Auxiliary class to store the result of a voting of
	 * (the biased) classifiers
	 */	
	private class Vote {
		public double agreement;
		public int votedClass;
		public Vote(double _agreement, int _votedClass) {
			agreement = _agreement;
			votedClass = _votedClass;
		}
	}
	
	/**
	 * Whether or not to produce additional output
	 */
	// boolean talkative = false;
	
	/**
	 * Add instances to arbiter, where biased classifiers agree,
	 * but on the wrong class?
	 */
	boolean addAgreeWrong = false;
	
	/**
	 * If true, use a 1-vs-all approach when handling multi-class problems 
	 */
	boolean oneVsAll = false;
	
	/**
	 * The fraction of classifiers that have to agree, or else the
	 * arbiter classifier will be asked.
	 */
	double minAgreement = 1.0;
	
	/**
	 * Base classifier for arbiter
	 */
	Classifier arbiterClassifier = new SMO();
	
	/**
	 * Unbiased classifier for stacking-like approach
	 */
	Classifier unbiasedClassifier = new SMO();
	
	/**
	 * The base classifier for the biased classifiers
	 */
	Classifier biasedClassifier = new UndersamplingBiasedClassifier();
	
	/**
	 * Whether or not to perform iterative weight adjustment
	 */
	boolean displayIRStats = true;
	
	/**
	 * Whether or not to assign lower weight to instances for the arbiter
	 * where the biased classifiers agreed instead of dropping them
	 * completely.
	 */
	boolean weightingInsteadOfDropping = false;
	
	/**
	 * The factor with which instances are weighted for the arbiter that
	 * would otherwise not make it into the set at all (see
	 * weightingInsteadOfDropping above)
	 */
	double weightFactorArbiter = 2;
	
	/**
	 * Whether or not a stacking-like approach is used to combine the
	 * predictions of the biased classifiers and the arbiter instead of
	 * a fixed rule (which is: ask arbiter if biased classifiers disagree).
	 */
	boolean stacking = false;
	
	/**
	 * The meta-classifier that is used for the stacking-like approach
	 */
	Classifier metaClassifier = new J48();
	
	/**
	 * The dataset for the meta classifier
	 */
	Instances metaInstances;
	
	public double classifyInstance(Instance i) throws Exception {
	    
	    if (stacking) {
	        
	        // If we are in stacking-like mode, create the "meta instance"
	        // by obtaining all predictions, then ask the meta classifier.
	        
	        double[] predictions = obtainAllPredictions(i);
	        Instance j = new Instance(1, predictions);
	        j.setDataset(metaInstances);
	        return metaClassifier.classifyInstance(j);
	    } else {
	        return classifyInstanceNonstacking(i);
	    }
	}
	
	public double classifyInstanceNonstacking(Instance i) throws Exception {
	    
	    // No stacking, use fixed rule to obtain final result

		// classify with all classifiers
		detailedResults = classifyWithBiasedClassifiers(i);

		detailedResults[clc.length] = arbiter[0].classifyInstance(i);
		detailedResults[clc.length+1] = -1;
		detailedResults[clc.length+2] = -1;

		int nClasses = clc.length;
		
		// if the agreement ratio is less than minAgreement,
		// ask the arbiter
		
		Vote vote = vote(detailedResults);
		
		// if (m_Debug) System.out.println("Agreement: " + vote.agreement);
		if (vote.agreement < minAgreement) {
			// ask arbiter
			// if (m_Debug) System.out.println("Asking arbiter");
			detailedResults[clc.length+1] = detailedResults[clc.length];
			return detailedResults[clc.length];
		} else {
			detailedResults[clc.length+2] = (double) vote.votedClass;
			if ((double) vote.votedClass != detailedResults[0] && m_Debug && nClasses == 2) {
				System.out.println("Whoops! vote.votedClass=" + vote.votedClass + ", cls=" + detailedResults[0]);
			}
			return (double) vote.votedClass;
		}
	}

	/**
	 * @param detailedResults The output of classifyWithBiasedClassifiers
	 * @return A Vote consisting of agreement ratio and the majority class
	 */
	private Vote vote(double[] detailedResults) {
		
		int nClasses = clc.length;

		if (oneVsAll) {
			
			// In the one-vs-all case, there is only either total agreement
			// (agreement=1) or disagreement (agreement=0). In the latter case,
			// votedClass is set to -1 (as it is useless anyway when
			// agreement=0). 
			
			// The classifiers have been trained on a dataset that was created
			// by a MakeIndicator filter, i.e. class index 0 means negative and
			// class index 1 means positive.
			
			int classification = -1;

			// loop all one-vs-all classifiers
			for (int n = 0; n < nClasses; n++) {
				int predictedClass = (int) detailedResults[n];
				if (predictedClass == 1) {	// classified as positive
				// if (predictedClass == 0) {	// classified as positive
					if (classification == -1) {
						classification = n;
					} else {
						// disagreement
						return new Vote(0, -1);
					}
				}
			}
			if (classification == -1) {
				// disagreement, ask arbiter
				return new Vote(0, -1);
			}
			// agreement
			detailedResults[clc.length+1] = (double) classification;
			return new Vote(1, classification);

		} else {
		
			// vote and find the class with max votes
				
			int[] nVotes = new int[nClasses]; 
			for (int n = 0; n < clc.length; n++) {
				nVotes[n] = 0;
			}
			int argmax = -1;
			int max = 0;
			for (int n = 0; n < nClasses; n++) {
				int predictedClass = (int) detailedResults[n];
				if (predictedClass != -1) {
					nVotes[predictedClass]++;
					if (nVotes[predictedClass] > max) {
						max = nVotes[predictedClass];
						argmax = predictedClass;
					}
				}
			}
			double agreement = (double) max / (double) nClasses;
			return new Vote(agreement, argmax);
		}
	}
	
	/**
	 * Classify a single instance with all biased classifiers
	 * @param i A single instance that will be classifierd
	 * @return An array of detailed results with room left for the arbiter
	 * @throws Exception
	 */
	private double[] classifyWithBiasedClassifiers(Instance i) throws Exception {

		// Reserve array for results of individual classifiers,
		// leaving room for arbiter.
		
		double[] detailedResults = new double[clc.length+3];

		// classify with all biased classifiers
		
		for (int n = 0; n < clc.length; n++) {
			try {
				detailedResults[n] = clc[n].classifyInstance(i);
			} catch (Exception e) {
				detailedResults[n] = -1;
				if (m_Debug) {
					System.err.println("Whoops! (Something strange happened in classifier " + n + ".)");
					e.printStackTrace();
				}
			}
		}
		return detailedResults;
	}
	
	Instances trainingSet;
	int nClasses;
	int nInstances;
	
	/* (non-Javadoc)
	 * @see weka.classifiers.Classifier#buildClassifier(weka.core.Instances)
	 */
	public void buildClassifier(Instances _trainingSet) throws Exception {
	    
	    // Build unbiased classifier, if necessary (ie. if in stacking mode)
	    if (stacking) unbiasedClassifier.buildClassifier(_trainingSet);
	    
	    trainingSet = _trainingSet;

		///////////////////////////////////////////////////////////////////
		// 	Create biased classifiers
		//
        
        if (!(biasedClassifier instanceof BiasedClassifier)) {
            throw new Exception("biasedClassifier is not a BiasedClassifier");
        }
		nClasses = trainingSet.numClasses();
		nInstances = trainingSet.numInstances();
		clc = Classifier.makeCopies(biasedClassifier, nClasses);
		for (int n = 0; n < nClasses; n++) {

			if (oneVsAll) {
				MakeIndicator classFilter = new MakeIndicator();
				classFilter.setValueIndex(n);
				classFilter.setNumeric(false);
				classFilter.setInputFormat(trainingSet);
				Instances newInsts = weka.filters.Filter.useFilter(trainingSet, classFilter);
				((BiasedClassifier) clc[n]).setPreferredClassIndex(1);
				if (separateBiases) {
				    if (n == 0) {
				        ((BiasedClassifier) clc[n]).setBiasStrength(bias0); 
				    } else {
				        ((BiasedClassifier) clc[n]).setBiasStrength(bias1); 
				    }
				}
				clc[n].buildClassifier(newInsts);				
			} else {
				((BiasedClassifier) clc[n]).setPreferredClassIndex(n);
				if (separateBiases) {
				    if (n == 0) {
				        ((BiasedClassifier) clc[n]).setBiasStrength(bias0); 
				    } else {
				        ((BiasedClassifier) clc[n]).setBiasStrength(bias1); 
				    }
				}
			    clc[n].buildClassifier(trainingSet);
			}
		}

		///////////////////////////////////////////////////////////////////
		// 	Evaluate biased classifiers
		//

		// Arbiter is trained on instances where biased classifiers disagree
		// (or agree, but on the wrong class)
		
		double[][] biasedResults = new double[nInstances][];
		Instances t = new Instances(trainingSet, 0);
		for (int m = 0; m < nInstances; m++) {
			Instance i = trainingSet.instance(m);
			int trueClass = (int) i.classValue();
			biasedResults[m] = classifyWithBiasedClassifiers(i);
			Vote v = vote(biasedResults[m]);
			
			if (v.agreement < minAgreement || ((v.votedClass != trueClass) && addAgreeWrong)) {
			    if (weightingInsteadOfDropping){
					Instance j = (Instance) i.copy();
					j.setWeight(i.weight() * weightFactorArbiter);
					t.add(j);
				} else {
					t.add(i);
				}
			} else if (weightingInsteadOfDropping) {
				Instance j = (Instance) i.copy();
				j.setWeight(i.weight() / weightFactorArbiter);
				t.add(j);
			}
		}
		
		///////////////////////////////////////////////////////////////////
		// Support for multiple rounds
		//

		if (rounds > 1) {
		    // When using weighting instead of dropping, it does not make ANY
		    // sense AT ALL to quit triskeling, if number of instances in the
		    // arbiter is the same as last round, because this will ALWAYS be
		    // the case if we don't drop any instances.
			if (t.numInstances() == arbiterInstancesInLastRound && !weightingInsteadOfDropping) {
				if (m_Debug) {
					System.out.println("Number of instances in dispute set same as in last round," +
						" suspending further rounds.");
				}
				arbiter = Classifier.makeCopies(arbiterClassifier, 1);
			} else if (t.numInstances() <= 3) {	// hardcoded constant!
				if (m_Debug && rounds > 1) {
					System.out.println("<= 3 instances in dispute set, suspending further rounds.");
				}
				arbiter = Classifier.makeCopies(arbiterClassifier, 1);
			} else {
				// everything fine, use another round of Triskel for arbiter				
				arbiter = Classifier.makeCopies(this, 1);
				((Triskel) arbiter[0]).setRounds(rounds - 1);
			    ((Triskel) arbiter[0]).xsetArbiterInstancesInLastRound(t.numInstances());
			    // avoid learning multiple meta-classifiers
			    ((Triskel) arbiter[0]).setStacking(false);
			}
		} else {
			arbiter = Classifier.makeCopies(arbiterClassifier, 1);
		}

		if (t.numInstances() <=3) {	// hardcoded constant!
			// If the biased classifiers agreed on every instance (!),
			// then train the arbiter on the complete training set.
			// (What else could we do!?)
			if (m_Debug) {
				System.out.println("Round " + rounds +
					": Too few instances in dispute set, building arbiter on full training set.");
			} 
			arbiter[0].buildClassifier(trainingSet);
		} else {
			// build arbiter normally
			if (m_Debug) {
				System.out.println("Round " + rounds +
					": Building arbiter on " + t.numInstances() + " instances.");
			}
			arbiter[0].buildClassifier(t);

			BufferedWriter bw = new BufferedWriter(new FileWriter("round" + rounds + "arbiter.arff"));
			bw.write(t.toString());
			bw.flush();
			bw.close();

		}
		
		///////////////////////////////////////////////////////////////////
		// Support for stacking-like combination of results
		// (instead of fixed rule)
		//
		
		if (stacking) {
		    double[][] predictions = obtainAllPredictions(trainingSet, biasedResults);
		    
		    // Create dataset for meta classifier
		    
		    Attribute classAttribute = trainingSet.classAttribute();
		    FastVector attInfo = new FastVector();
		    for (int n = 0; n < predictions[0].length - 3; n++) {
		        
		        // !! The following line required a slight change in Attribute:
		        // !! The copy(String) method had to be made public!
		        
		        String name = "round" + ((n / nClasses) + 1) + "biased" + (n % nClasses);  
		        attInfo.addElement(classAttribute.copy(name));
		    }
		    attInfo.addElement(classAttribute.copy("arbiter"));
		    attInfo.addElement(classAttribute.copy("unbiased")); // new
		    attInfo.addElement(classAttribute.copy("class"));
		    metaInstances = new Instances("meta", attInfo, nInstances);
		    metaInstances.setClassIndex(metaInstances.numAttributes() - 1);
		    
		    // Create instances for meta classifier
		    
		    for (int n = 0; n < nInstances; n++) {
		        metaInstances.add(new Instance(1, predictions[n]));
		    }
		    
		    // Train meta classifier
		    
		    metaClassifier.buildClassifier(metaInstances);
		}
		
	}
	
	protected double[][] obtainAllPredictions(Instances instances,
	        double[][] biasedPredictions) throws Exception {
	
	    int nInstances = instances.numInstances();
	    double[][] predictions = new double[nInstances][];

	    // For efficiency, re-use the predictions from the biased classifiers
	    // that were already made

	    if (arbiter[0] instanceof Triskel) {
	        for (int n = 0; n < nInstances; n++) {
	            double[] arbiterPredictions = ((Triskel) arbiter[0]).obtainAllPredictions(instances.instance(n));
		        predictions[n] = new double[nClasses + arbiterPredictions.length];
	            System.arraycopy(biasedPredictions[n], 0, predictions[n], 0, nClasses);
	            System.arraycopy(arbiterPredictions, 0, predictions[n], nClasses, arbiterPredictions.length);
	            // predictions[n][predictions[n].length-1] = instances.instance(n).classValue(); 
	        }
	    } else {
		    for (int n = 0; n < nInstances; n++) {
		        // predictions[n] = new double[nClasses + 2];
		        predictions[n] = new double[nClasses + 3]; // unbiased
		        System.arraycopy(biasedPredictions[n], 0, predictions[n], 0, nClasses);
		        predictions[n][nClasses] = arbiter[0].classifyInstance(instances.instance(n));
		        if (stacking) {
		            predictions[n][nClasses + 1] = unbiasedClassifier.classifyInstance(instances.instance(n));
		        } else {
		            predictions[n][nClasses + 1] = 0;
		        }
	            predictions[n][predictions[n].length-1] = instances.instance(n).classValue(); 
		    }
	    }
	    return predictions;
	}
	
	public double[] obtainAllPredictions(Instance i) throws Exception {

	    // Array contains predictions in the following order:
	    // First all predictions of the biased classifier
	    // Then, if the arbiter is itself a Triskel, the return values of
	    // this method invoked on the arbiter,
	    // or, if the arbiter is an ordinary classifier, the prediction
	    // of the arbiter.
	    // Finally, the class value taken from i is appended.
	    // Eventually, this means that the return value will consist of
	    // an array containing all predictions of all biased classifiers
	    // from *all* rounds plus the prediction of the inner arbiter
	    // plus the class value from the instance.

	    double[] predictions;
        double[] biasedPredictions = classifyWithBiasedClassifiers(i);
	    if (arbiter[0] instanceof Triskel) {
	        double[] arbiterPredictions =
	            ((Triskel) arbiter[0]).obtainAllPredictions(i);
	        predictions = new double[nClasses + arbiterPredictions.length];
            System.arraycopy(biasedPredictions, 0, predictions, 0, nClasses);
            System.arraycopy(arbiterPredictions, 0, predictions, nClasses, arbiterPredictions.length);
	    } else {
	        // predictions = new double[nClasses + 2];
	        predictions = new double[nClasses + 3]; // unbiased
            System.arraycopy(biasedPredictions, 0, predictions, 0, nClasses);
            predictions[nClasses] = arbiter[0].classifyInstance(i);
            if (stacking) {
                predictions[nClasses+1] = unbiasedClassifier.classifyInstance(i);
            } else {
                predictions[nClasses+1] = 0;
            }
            predictions[nClasses+2] = i.classValue();
	    }
	    return predictions;
	}

	/* (non-Javadoc)
	 * @see weka.core.OptionHandler#listOptions()
	 */
	public Enumeration listOptions() {
		Vector options = new Vector();
		options.add(
				new Option("\tClassifier for arbiter" +
						"(default SMO)", "W", 1, "-W <classifier>"));
		options.add(
				new Option("\tBiased classifiers" +
						"(default UndersamplingBiasedClassifier)", "B", 1, "-B <classifier>"));
		options.add(
				new Option("\tToggle debug output (0 off or 1 on, default 0)",
						"D", 1, "-D <int>"));
		options.add(
				new Option("\tWhether or not to add instances where the biased classifiers " +
					"agreed on the wrong class to the arbiter (0 off or 1 on, default 0)",
						"Q", 1, "-A <int>"));
		options.add(
				new Option("\tUse 1-vs-All mode for multiclass classification (default 0)",
						"C", 1, "-C <int>"));
		options.add(
				new Option("\tMinimum agreement of classifiers for not asking arbiter (default 1.0)",
						"A", 1, "-A <double>"));
		options.add(
				new Option("\tSpecify the number of rounds for this algorithm. If > 1, the arbiter " +
						"classifier will be replaced with another Triskel.",
						"R", 1, "-R <int>"));
		options.add(
				new Option("\tWhether or not to use weighting instead of dropping instances when " +
					"training the arbiter. (default 1)",
					"U", 1, "-U <int>"));
		options.add(
				new Option("\tWhether or not to use stacking-like combination of classifiers " +
					"instead of fixed rule. (default 0)",
					"S", 1, "-S <int>"));
		options.add(
				new Option("\tMeta classifier for -S 1" +
						"(default J48)", "M", 1, "-M <classifier>"));
		options.add(
				new Option("\tWhen weighting for the arbiter instead of dropping instances, use this " +
					"as a factor for weighting. The current weight is multiplied with this number. (default 0.5)",
					"V", 1, "-V <double>"));
		return options.elements();
	}

	/* (non-Javadoc)
	 * @see weka.core.OptionHandler#setOptions(java.lang.String[])
	 */
	public void setOptions(String[] options) throws Exception {
		
		String arbiterString = Utils.getOption('W', options);
	    String [] arbiterSpec = Utils.splitOptions(arbiterString);
	    String arbiterName;
	    if (arbiterSpec.length > 0) {
	        arbiterName = arbiterSpec[0];
	        arbiterSpec[0] = "";
	        arbiterClassifier = Classifier.forName(arbiterName, arbiterSpec);
	    }

	    String biasedString = Utils.getOption('B', options);
	    String [] biasedSpec = Utils.splitOptions(biasedString);
	    String biasedName;
	    if (biasedSpec.length > 0) {
	        biasedName = biasedSpec[0];
	        biasedSpec[0] = "";
	        biasedClassifier = Classifier.forName(biasedName, biasedSpec);
	    }
		
		String addAgreeWrongOption = Utils.getOption('Q', options);
		if (addAgreeWrongOption.length() != 0) {
			int addAgreeWrongOptionInt = Integer.parseInt(addAgreeWrongOption);
			if (addAgreeWrongOptionInt == 1) addAgreeWrong = true;
		}

		String oneVsAllOption = Utils.getOption('C', options);
		if (oneVsAllOption.length() != 0) {
			int oneVsAllOptionInt = Integer.parseInt(oneVsAllOption);
			if (oneVsAllOptionInt == 1) oneVsAll = true;
		}
		
		String minAgreeOption = Utils.getOption('A', options);
		if (minAgreeOption.length() != 0) {
			minAgreement = Double.parseDouble(minAgreeOption);
		}

		String roundsOption = Utils.getOption('R', options);
		if (roundsOption.length() != 0) {
			rounds = Integer.parseInt(roundsOption);
		}

		String weightingInsteadOption = Utils.getOption('U', options);
		if (weightingInsteadOption.length() != 0) {
			int weightingInsteadOptionInt = Integer.parseInt(weightingInsteadOption);
			if (weightingInsteadOptionInt == 1) weightingInsteadOfDropping = true;
		}

		String weightFactorOption = Utils.getOption('V', options);
		if (weightFactorOption.length() != 0) {
			weightFactorArbiter = Double.parseDouble(weightFactorOption);
		}

		String stackingOption = Utils.getOption('S', options);
		if (stackingOption.length() != 0) {
			int stackingOptionInt = Integer.parseInt(stackingOption);
			if (stackingOptionInt == 1) stacking = true;
		}

		String metaString = Utils.getOption('M', options);
	    String [] metaSpec = Utils.splitOptions(metaString);
	    String metaName;
	    if (metaSpec.length > 0) {
	        metaName = metaSpec[0];
	        metaSpec[0] = "";
	        metaClassifier = Classifier.forName(metaName, metaSpec);
	    }
	}

	/* (non-Javadoc)
	 * @see weka.core.OptionHandler#getOptions()
	 */
	public String[] getOptions() {

	    String [] options = new String [20];

		int c = 0;
		
		options[c++] = "-W"; options[c++] =
		    arbiterClassifier.getClass().getName() +
			" " + Utils.joinOptions(arbiterClassifier.getOptions());
		options[c++] = "-B"; options[c++] =
		    biasedClassifier.getClass().getName() +
		    " " + Utils.joinOptions(biasedClassifier.getOptions());
		options[c++] = "-A"; options[c++] = "" + minAgreement;
		options[c++] = "-R"; options[c++] = "" + rounds;
		if (addAgreeWrong) {
			options[c++] = "-Q"; options[c++] = "1";
		} else {
			options[c++] = "-Q"; options[c++] = "0";
		}
		if (oneVsAll) {
			options[c++] = "-C"; options[c++] = "1";
		} else {
			options[c++] = "-C"; options[c++] = "0";
		}
		if (weightingInsteadOfDropping) {
			options[c++] = "-U"; options[c++] = "1";
		} else {
			options[c++] = "-U"; options[c++] = "0";
		}
		options[c++] = "-V"; options[c++] = "" + weightFactorArbiter;
		if (stacking) {
			options[c++] = "-S"; options[c++] = "1";
		} else {
			options[c++] = "-S"; options[c++] = "0";
		}
		options[c++] = "-M"; options[c++] =
		    metaClassifier.getClass().getName() +
		    " " + Utils.joinOptions(metaClassifier.getOptions());
	    while (c < options.length) {
	      options[c++] = "";
	    }

		return options;
	}
	
	/**
	 * Main method for testing this class.
	 * Note that if you call the classifier this way, the default settings
	 * are SMO and random undersampling (as opposed to NaiveBayes and
	 * overweighting).
	 */
	public static void main(String[] argv) {

	  Triskel scheme;

	  try {
		scheme = new Triskel();
		scheme.setArbiterClassifier(new SMO());
		scheme.setRounds(2);
		System.out.println(Evaluation.evaluateModel(scheme, argv));
	  } catch (Exception e) {
		e.printStackTrace();
		System.err.println(e.getMessage());
	  }
	}

	public Classifier getArbiterClassifier() {
		return arbiterClassifier;
	}

	public void setArbiterClassifier(Classifier classifier) {
		arbiterClassifier = classifier;
	}

	public String seedTipText() {
		return "Used as seed for random number generator if random undersampling is selected.";
	}

	public double[] getDetailedResults() {
		return detailedResults;
	}

	public String oversamplingWeightTipText() {
		return "Weight that is assigned to positive instances in oversampling mode";
	}

	public String globalInfo() {
		return "An ensemble method that trains specialised classifiers on " +
				"subsets of instances so that effectively the feature space is " +
				"split into sections. (Andreas Hess, University College Dublin," +
				" <andreas.hess@ucd.ie>)";
	}

	/**
	 * @return
	 */
	public double getMinAgreement() {
		return minAgreement;
	}

	/**
	 * @return
	 */
	public boolean isOneVsAll() {
		return oneVsAll;
	}

	/**
	 * @param d
	 */
	public void setMinAgreement(double d) {
		minAgreement = d;
	}

	/**
	 * @param b
	 */
	public void setOneVsAll(boolean b) {
		oneVsAll = b;
	}

	/**
	 * @return
	 */
	public int getRounds() {
		return rounds;
	}

	/**
	 * @param i
	 */
	public void setRounds(int i) {
		rounds = i;
	}

	/**
	 * @return
	 */
	public boolean isAddAgreeWrong() {
		return addAgreeWrong;
	}

	/**
	 * @param b
	 */
	public void setAddAgreeWrong(boolean b) {
		addAgreeWrong = b;
	}

	/**
	 * @param i
	 */
	public void xsetArbiterInstancesInLastRound(int i) {
		arbiterInstancesInLastRound = i;
	}
	
	/**
	 * @return
	 */
	public boolean isDisplayIRStats() {
		return displayIRStats;
	}

	/**
	 * @param b
	 */
	public void setDisplayIRStats(boolean b) {
		displayIRStats = b;
	}

	/**
	 * @return
	 */
	public double getWeightFactorArbiter() {
		return weightFactorArbiter;
	}

	/**
	 * @return
	 */
	public boolean isWeightingInsteadOfDropping() {
		return weightingInsteadOfDropping;
	}

	/**
	 * @param d
	 */
	public void setWeightFactorArbiter(double d) {
		weightFactorArbiter = d;
	}

	/**
	 * @param b
	 */
	public void setWeightingInsteadOfDropping(boolean b) {
		weightingInsteadOfDropping = b;
	}

    /**
     * @return Returns the biasedClassifier.
     */
    public Classifier getBiasedClassifier() {
        return biasedClassifier;
    }
    /**
     * @param biasedClassifier The biasedClassifier to set.
     */
    public void setBiasedClassifier(Classifier biasedClassifier) {
        this.biasedClassifier = biasedClassifier;
    }
    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString() {
        
        StringBuffer output = new StringBuffer();
        
        try {
            for (int n = 0; n < nClasses; n++) {
                output.append("*** Biased classifier for class " + n + ":\n\n");
                output.append(clc[n].toString());
                output.append("\n\n");
            }
        } catch (RuntimeException e1) {
            output.append(e1.toString());
        }
        
        try {
            output.append("*** Arbiter:\n\n");
            output.append(arbiter[0].toString());
        } catch (RuntimeException e2) {
            output.append(e2.toString());
        }
        
        if (m_Debug) {
            output.append("+++ Instances in arbiter:\n\n");
        }
        
        try {
            if (displayIRStats) {
                
                output.append("\n\n*** Training Set IR statistics\n\n");
                
                // number of TP/TN/FP/FN classified instances per classifier
                // In a multi-class setting, values are wrt. to the class that
                // we're biasing towards.
                double[] nTruePositives = new double[nClasses];
                double[] nTrueNegatives = new double[nClasses];
                double[] nFalsePositives = new double[nClasses];
                double[] nFalseNegatives = new double[nClasses];

                int arbiterCorrect = 0;
                
                // perform a training-set validation
                for (int m = 0; m < nInstances; m++) {
                    double[] predictions = classifyWithBiasedClassifiers(trainingSet.instance(m));
                    predictions[nClasses] = arbiter[0].classifyInstance(trainingSet.instance(m));                  
                    double trueClass = trainingSet.instance(m).classValue(); 
                    for (int n = 0; n < nClasses; n++) {
                        
                        // "translation" for oneVsAll mode
                        if (oneVsAll) {
                            if (predictions[n] == 1) {
                                predictions[n] = n;
                            } else {
                                if (trueClass != n) {
                                    predictions[n] = trueClass;
                                } else {
                                    predictions[n] = -1;
                                }
                            }
                            // predictions[n] = (predictions[n] == 1 ? (double) n : -1);
                            // predictions[n] = (predictions[n] == 0 ? (double) n : -1);
                        }
                        
                        if (predictions[n] == trueClass) {
                            if (trueClass == n) {
                                // classified correctly as positive
                                nTruePositives[n]++;
                            } else {
                                // classified correctly as negative
                                nTrueNegatives[n]++;
                            }
                        } else {
                            if (trueClass == n) {
                                // classified incorrectly, but is indeed positive
                                // -> false negative
                                nFalseNegatives[n]++;
                            } else {
                                // classified incorrectly, but is indeed negative
                                // -> false positive
                                nFalsePositives[n]++;
                            }
                        }
                    } // for 0..nClasses
                    
                    if (predictions[nClasses] == trueClass) {
                        arbiterCorrect++;
                    }
                    
                } // for 0..nInstances
                
                // Show evaluation
                double[] precisionPositiveClass = new double[nClasses];
                double[] precisionNegativeClass = new double[nClasses];
                double[] recallPositiveClass = new double[nClasses];
                double[] recallNegativeClass = new double[nClasses];
                double[] accuracy = new double[nClasses];
                for (int n = 0; n < nClasses; n++) {
                    precisionPositiveClass[n] = nTruePositives[n] / (nTruePositives[n] + nFalsePositives[n]);
                    precisionNegativeClass[n] = nTrueNegatives[n] / (nTrueNegatives[n] + nFalseNegatives[n]);
                    recallPositiveClass[n] = nTruePositives[n] / (nTruePositives[n] + nFalseNegatives[n]);
                    recallNegativeClass[n] = nTrueNegatives[n] / (nTrueNegatives[n] + nFalsePositives[n]);
                    accuracy[n] = (nTruePositives[n] + nTrueNegatives[n]) /
                    (nTruePositives[n] + nTrueNegatives[n] + nFalsePositives[n] + nFalseNegatives[n]);
                    output.append("\n\nIR stats for biased classifier " + n + ":\n\n");
                    output.append("TP: " + nTruePositives[n] + " FP: " + nFalsePositives[n] +
                            " TN: " + nTrueNegatives[n] + " FN: " + nFalseNegatives[n] + "\n");
                    output.append("Positive Class Precision: " + precisionPositiveClass[n] +
                            "   Recall: " + recallPositiveClass[n] + "\n");
                    output.append("Negative Class Precision: " + precisionNegativeClass[n] +
                            "   Recall: " + recallNegativeClass[n] + "\n");
                    output.append("Accuracy: " + accuracy[n] + "\n");
                }
                
                output.append("\nArbiter accuracy: " + arbiterCorrect + " / " + nInstances +
                        " = " + (double) arbiterCorrect / (double) nInstances + "\n");
            }
        } catch (Exception e) {
            output.append(e.toString());
        }
        
        if (stacking) {
            try {
                output.append("*** Meta Classifier:\n\n");
                output.append(metaClassifier.toString());
            } catch (RuntimeException e3) {
                output.append(e3.toString());
            }
        }
        
        return output.toString();
   }
    public boolean isStacking() {
        return stacking;
    }
    public void setStacking(boolean stacking) {
        this.stacking = stacking;
    }
    public Classifier getMetaClassifier() {
        return metaClassifier;
    }
    public void setMetaClassifier(Classifier metaClassifier) {
        this.metaClassifier = metaClassifier;
    }
    public double getBias0() {
        return bias0;
    }
    public void setBias0(double bias0) {
        this.bias0 = bias0;
    }
    public double getBias1() {
        return bias1;
    }
    public void setBias1(double bias1) {
        this.bias1 = bias1;
    }
    public boolean isSeparateBiases() {
        return separateBiases;
    }
    public void setSeparateBiases(boolean separateBiases) {
        this.separateBiases = separateBiases;
    }
    public Classifier getUnbiasedClassifier() {
        return unbiasedClassifier;
    }
    public void setUnbiasedClassifier(Classifier unbiasedClassifier) {
        this.unbiasedClassifier = unbiasedClassifier;
    }
}
