/*
 * Created on 02.10.2004
 */
package weka.classifiers.ucd.ensemble;

import java.io.Serializable;
import java.util.Enumeration;
import java.util.Random;
import java.util.Vector;

import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.classifiers.bayes.NaiveBayes;
import weka.classifiers.functions.SMO;
import weka.classifiers.ucd.biased.BiasedClassifier;
import weka.core.*;
import weka.filters.unsupervised.attribute.MakeIndicator;

/**
 * @author andreas
 */
public class TriskelA extends Classifier implements OptionHandler, Serializable {
	
	/**
	 * some additional statistics
	 * element 0-(nClasses-1): individual classifiers
	 * element nClasses: arbiter
	 * element nClasses+1: arbiter (when actually used)
	 * element nClasses+2: -1 if disagree, = element 0 otherwise 
	 */
	double detailedResults[];
	
	/**
	 * Instances for subclassifiers
	 */
	Instances[] lc;
	
	/**
	 * Subclassifiers
	 */
	Classifier[] clc;
	
	/**
	 * Subclassifier
	 */
	Classifier[] arbiter;
	
	/**
	 * The mode of operation
	 */
	int method = 2;
	
	/**
	 * If rounds > 1, then the arbiter will be replaced with another
	 * Triskel with the same parameters as this one except
	 * that rounds will be decreased by 1.
	 */
	int rounds = 1;

	/**
	 * The number of instances the arbiter was trained on in the last round.
	 * If the arbiter in the next round is trained on the same number of
	 * instances (ie. the biased classifiers disagreed on every instance!)
	 * then further rounds are suspended.
	 */	
	int arbiterInstancesInLastRound = -1;

	/**
	 * Auxiliary class to store the result of a voting of
	 * (the biased) classifiers
	 */	
	private class Vote {
		public double agreement;
		public int votedClass;
		public Vote(double _agreement, int _votedClass) {
			agreement = _agreement;
			votedClass = _votedClass;
		}
	}
	
	public static final int METHOD_UNDERSAMPLING = 1;
	public static final int METHOD_OVERSAMPLING = 2;
	public static final int METHOD_TRAININGERROR = 3;
	public static final int METHOD_BIAS = 4;
	public static final Tag[] TAGS_METHOD = {
			new Tag(METHOD_UNDERSAMPLING, "Random undersampling"),
			new Tag(METHOD_OVERSAMPLING, "Weighted oversampling"),
			new Tag(METHOD_TRAININGERROR, "Training-error based"),
			new Tag(METHOD_BIAS, "Bias base classifier directly")
	};

	/**
	 * Whether or not to produce additional output
	 */
	boolean talkative = false;
	
	/**
	 * Add instances to arbiter, where biased classifiers agree,
	 * but on the wrong class?
	 */
	boolean addAgreeWrong = false;
	
	/**
	 * If this is true, all instances that are inserted into
	 * the arbiter classifier get a weight of 4.
	 * 
	 * As stupid as it sounds, but this can boost accuracy by
	 * about 5% !!
	 * 
	 * However, by default, this option is off, so that we cannot
	 * be accused of cheating.
	 */
	boolean smoCheat = false;
	
	/**
	 * Whether or not to use log based weighting instead of fixed weights/probabilities
	 */
	boolean logBasedWeighting = false;
	
	/**
	 * If true, use a 1-vs-all approach when handling multi-class problems 
	 */
	boolean oneVsAll = false;
	
	/**
	 * The fraction of classifiers that have to agree, or else the
	 * arbiter classifier will be asked.
	 */
	double minAgreement = 1.0;
	
	/**
	 * per-class probabilities for undersampling
	 */
	double undersamplingProbability = 0.1;
	
	/**
	 * weight-multiplicator for oversampling
	 */
	double oversamplingWeight = 3; // good in general
	
	/**
	 * Base classifier class name (default Naive Bayes)
	 */
	// String classifier = "weka.classifiers.bayes.NaiveBayes";
	Classifier classifier = new NaiveBayes();
	
	/**
	 * Options for base classifier
	 */
	// String[] classifierOptions;
	
	/**
	 * Random number seed (for random undersampling)
	 */
	int seed = 1;
	
	/**
	 * Number of iterations when adjusting the weights for the biased classifier
	 * by means of validation on training set
	 */
	int weightAdjustmentIterations = 10;

	/**
	 * Whether or not to perform iterative weight adjustment
	 */
	boolean adjustWeights = true;
	
	/**
	 * Whether or not to assign lower weight to instances for the arbiter
	 * where the biased classifiers agreed instead of dropping them
	 * completely.
	 */
	boolean weightingInsteadOfDropping = false;
	
	/**
	 * The factor with which instances are weighted for the arbiter that
	 * would otherwise not make it into the set at all (see
	 * weightingInsteadOfDropping above)
	 */
	double weightFactorArbiter = 2;
	
	public double classifyInstance(Instance i) throws Exception {

		// classify with all classifiers
		detailedResults = classifyWithBiasedClassifiers(i);

		detailedResults[clc.length] = arbiter[0].classifyInstance(i);
		detailedResults[clc.length+1] = -1;
		detailedResults[clc.length+2] = -1;

		int nClasses = clc.length;
		
		// if the agreement ratio is less than minAgreement,
		// ask the arbiter
		
		Vote vote = vote(detailedResults);
		
		// if (talkative) System.out.println("Agreement: " + vote.agreement);
		if (vote.agreement < minAgreement) {
			// ask arbiter
			// if (talkative) System.out.println("Asking arbiter");
			detailedResults[clc.length+1] = detailedResults[clc.length];
			return detailedResults[clc.length];
		} else {
			detailedResults[clc.length+2] = (double) vote.votedClass;
//			if ((double) vote.votedClass != detailedResults[0] && talkative && nClasses == 2) {
//				System.out.println("Whoops! vote.votedClass=" + vote.votedClass + ", cls=" + detailedResults[0]);
//			}
			return (double) vote.votedClass;
		}
	}

	/**
	 * @param detailedResults The output of classifyWithBiasedClassifiers
	 * @return A Vote consisting of agreement ratio and the majority class
	 */
	private Vote vote(double[] detailedResults) {
		
		int nClasses = clc.length;

		if (oneVsAll) {
			
			// In the one-vs-all case, there is only either total agreement
			// (agreement=1) or disagreement (agreement=0). In the latter case,
			// votedClass is set to -1 (as it is useless anyway when
			// agreement=0). 
			
			// The classifiers have been trained on a dataset that was created
			// by a MakeIndicator filter, i.e. class index 0 means negative and
			// class index 1 means positive.
			
			int classification = -1;

			// loop all one-vs-all classifiers
			for (int n = 0; n < nClasses; n++) {
				int predictedClass = (int) detailedResults[n];
				if (predictedClass == 1) {	// classified as positive
					if (classification == -1) {
						classification = n;
					} else {
						// disagreement
						return new Vote(0, -1);
					}
				}
			}
			if (classification == -1) {
				// disagreement, ask arbiter
				return new Vote(0, -1);
			}
			// agreement
			detailedResults[clc.length+1] = (double) classification;
			return new Vote(1, classification);

		} else {
		
			// vote and find the class with max votes
				
			int[] nVotes = new int[nClasses]; 
			for (int n = 0; n < clc.length; n++) {
				nVotes[n] = 0;
			}
			int argmax = -1;
			int max = 0;
			for (int n = 0; n < nClasses; n++) {
				int predictedClass = (int) detailedResults[n];
				if (predictedClass != -1) {
					nVotes[predictedClass]++;
					if (nVotes[predictedClass] > max) {
						max = nVotes[predictedClass];
						argmax = predictedClass;
					}
				}
			}
			double agreement = (double) max / (double) nClasses;
			return new Vote(agreement, argmax);
		}
	}
	
	/**
	 * Classify a single instance with all biased classifiers
	 * @param i A single instance that will be classifierd
	 * @return An array of detailed results with room left for the arbiter
	 * @throws Exception
	 */
	private double[] classifyWithBiasedClassifiers(Instance i) throws Exception {

		// Reserve array for results of individual classifiers,
		// leaving room for arbiter.
		
		double[] detailedResults = new double[clc.length+3];

		// classify with all biased classifiers
		
		for (int n = 0; n < clc.length; n++) {
			try {
				detailedResults[n] = clc[n].classifyInstance(i);
			} catch (Exception e) {
				detailedResults[n] = -1;
				if (talkative) {
					System.err.println("Whoops! (Something strange happened in classifier " + n + ".)");
					e.printStackTrace();
				}
			}
		}
		return detailedResults;
	}
	
	/* (non-Javadoc)
	 * @see weka.classifiers.Classifier#buildClassifier(weka.core.Instances)
	 */
	public void buildClassifier(Instances trainingSet) throws Exception {
		Random rnd = new Random(seed); // seed
		int nClasses = trainingSet.numClasses();
		int nInstances = trainingSet.numInstances();
		lc = new Instances[nClasses];
		clc = new Classifier[nClasses];

		int[] nInstancesPerClass = new int[nClasses];

		if (logBasedWeighting) {
			for (int m = 0; m < nInstances; m++) {
				Instance i = trainingSet.instance(m);
				nInstancesPerClass[(int) i.classValue()]++;
			}
		}
		
		int wa = 0;
//		boolean weightAdjustmentDone = false;
		
		///////////////////////////////////////////////////////////////////
		// 	Self-adjusting weights
		//  TODO The idea is that by doing some rounds of training set
		//  evaluation, the weights should automatically be set for each
		//  class so that the biased classifiers achieve high recall on the
		//  class that they are biased towards and high recall on the other
		//  class (i.e. false negatives should be avoided, where as false
		//  positives don't matter that much.)

		// do {
		
		switch (method) {
			
			///////////////////////////////////////////////////////////////////
			// 	METHOD_TRAININGERROR
			//
			
			case METHOD_TRAININGERROR:

				// Create a simple classifier first
				// This "pre-classifier" is just for finding out the
				// training set mistakes. With this information, biased
				// classifiers for each class are created with all instances *except*
				// the training set mistakes that belong to other classes
				
				Classifier preClassifier[] = Classifier.makeCopies(classifier, 1);
				preClassifier[0].buildClassifier(trainingSet);
				
				// initialize datasets for biased classifier
		
				for (int n = 0; n < nClasses; n++) {
					lc[n] = new Instances(trainingSet, 0);
				}
		
				// Compute training set errors
				
				int trainingErrors = 0;
				
				// This strategy seems to work best, at least with SVM and
				// the diabetes dataset:
				// For each biased classifier:
				// Instances of that class with training error are omitted
				
				for (int m = 0; m < nInstances; m++) {
					Instance i = trainingSet.instance(m);
					int trueClass = (int) i.classValue();
					int preClassifierSays = (int) preClassifier[0].classifyInstance(i);
					
					// add to classifier for true class only if correct
					
					if (trueClass == preClassifierSays) {
						lc[trueClass].add(i);
					} else {
						trainingErrors++;
					}
					
					// add to all classifiers except for the true class
					
					for (int n = 0; n < nClasses; n++) {
						if (n != trueClass) {
							lc[n].add(i);
						}
					}
				}
						
				if (talkative) System.out.println("Training errors: " + trainingErrors);
				
				break;

			///////////////////////////////////////////////////////////////////
			// 	METHOD_OVERSAMPLING
			//

			case METHOD_OVERSAMPLING:
			
				// Create a biased classifier for each class
				// by overweighting positive instances
	
				for (int n = 0; n < nClasses; n++) {
					
					double trueWeight = oversamplingWeight;
					
					if (logBasedWeighting) {
						int sum = 0;
						for (int a = 0; a < nClasses; a++) {
							if (a != n) {
								sum += nInstancesPerClass[a];
							}
						}

//						trueWeight = Math.log(sum);

						// TODO Note that this is not log-based...
						// This option is named "log based weighting"
						// for historic reasons. (See the commented
						// line above.)
						
//						trueWeight = oversamplingWeight *
//							(double) sum / (double) (nInstancesPerClass[n] + sum);

						trueWeight = oversamplingWeight /
							((double) nInstancesPerClass[n] / (double) sum);
						
						// For very large classes we have to make sure
						// that we don't underweight them
						if (trueWeight < 1) trueWeight = 1;
							
						if (talkative) {
							System.out.println("Class " + n + ", " + nInstancesPerClass[n] +
								" instances, adjusted weight: " + trueWeight);
						}
					}
					
					lc[n] = new Instances(trainingSet, 0);
					for (int m = 0; m < nInstances; m++) {
						Instance i = trainingSet.instance(m); 
						if ((trainingSet.instance(m).classValue() == (double) n)) {
							i = (Instance) i.copy();
							i.setWeight(trueWeight * i.weight()); // incorporate previous weights!
						}
						lc[n].add(i);
					}
				}
				break;
				
				
			///////////////////////////////////////////////////////////////////
			// 	METHOD_UNDERSAMPLING
			//

			case METHOD_UNDERSAMPLING:
	
				// Create a biased classifier for each class
				// by omitting negative instances
				// (negative instances are inserted with probability p)
				
				for (int n = 0; n < nClasses; n++) {
	
					double trueWeight = undersamplingProbability;
					if (logBasedWeighting) {
						int sum = 0;
						for (int a = 0; a < nClasses; a++) {
							if (a != n) {
								sum += nInstancesPerClass[a];
							}
						}

//						trueWeight = 1.0 / Math.log(nInstancesPerClass[n]);

						// TODO Note that this is not log-based...
						// This option is named "log based weighting"
						// for historic reasons. (See the commented
						// line above.)
						
						trueWeight = undersamplingProbability /
							((double) nInstancesPerClass[n] / (double) sum);

						if (talkative) {
							System.out.println("Class " + n + ", " + nInstancesPerClass[n] +
								" instances, adjusted weight: " + trueWeight);
						}
					}
					
					// Changed this method, so that for a 0.1 probability
					// every 10th instance is inserted instead of inserting
					// every instance with a chance of 0.1. This is not really
					// random any more, but the effect should be the same,
					// because when doing cross-validation, the splits are
					// randomised in advance. The advantage is that for low
					// values of undersamplingProbability, this method here is
					// guaranteed to include at least 1 instance. ;-)
					
					double counter = 1.0 + trueWeight;

					lc[n] = new Instances(trainingSet, 0);
					for (int m = 0; m < nInstances; m++) {
						Instance i = trainingSet.instance(m); 
						if ((i.classValue() != (double) n)) {
//							if (rnd.nextDouble() < trueWeight) {
							if (counter > 1) {
								// insert a copy of the instance
								lc[n].add(i);
								counter--;
							}
							counter += trueWeight;
						} else {
							// insert a copy of the instance
							lc[n].add(i);
						}
					}
				}
				break;
				
			///////////////////////////////////////////////////////////////////
			// 	METHOD_BIAS
			//

			case METHOD_BIAS:
			
				if ((!(classifier instanceof BiasedClassifier)) || nClasses > 2) {
					throw new Exception("Cannot set base classifier bias or problem not binary");
				}
			
				for (int n = 0; n < nClasses; n++) {
					lc[n] = new Instances(trainingSet, 0);
					for (int m = 0; m < nInstances; m++) {
						Instance i = trainingSet.instance(m); 
						lc[n].add(i);
					}
				}
			
			
				break;
				
				
			default:
				throw new Exception("Mode not correctly specified!");
		}

		// Create a biased classifier for each class
			
		clc = Classifier.makeCopies(classifier, nClasses);
		for (int n = 0; n < nClasses; n++) {
			
			// if one-vs-all, re-label instances for the biased classifiers
			
			if (oneVsAll) {
				MakeIndicator classFilter = new MakeIndicator();
				classFilter.setValueIndex(n);
				classFilter.setNumeric(false);
				classFilter.setInputFormat(lc[n]);
				Instances newInsts = weka.filters.Filter.useFilter(lc[n], classFilter);
				lc[n] = newInsts;				
			}
			
			clc[n].buildClassifier(lc[n]);
			
			if (method == METHOD_BIAS) {
				((BiasedClassifier) clc[0]).setBiasStrength(-1);
				((BiasedClassifier) clc[1]).setBiasStrength(1);
			}
		}
		
		///////////////////////////////////////////////////////////////////
		// 	Self-adjusting weights
		//  TODO At the moment, this just prints some statistics on the
		//  performance of the biased classifiers on the training set,
		//  but does not update the weights.

		if (adjustWeights) {
			
			// number of TP/TN/FP/FN classified instances per classifier
			// In a multi-class setting, values are wrt. to the class that
			// we're biasing towards.
			double[] nTruePositives = new double[nClasses];
			double[] nTrueNegatives = new double[nClasses];
			double[] nFalsePositives = new double[nClasses];
			double[] nFalseNegatives = new double[nClasses];
			
			// perform a training-set validation of the biased classifiers
			for (int m = 0; m < nInstances; m++) {
				double[] predictions = classifyWithBiasedClassifiers(trainingSet.instance(m));
				double trueClass = trainingSet.instance(m).classValue(); 
				for (int n = 0; n < nClasses; n++) {
					
					// "translation" for oneVsAll mode
					if (oneVsAll) {
						predictions[n] = predictions[n] == 1 ? (double) n : -1;
					}
					
					if (predictions[n] == trueClass) {
						if (trueClass == n) {
							// classified correctly as positive
							nTruePositives[n]++;
						} else {
							// classified correctly as negative
							nTrueNegatives[n]++;
						}
					} else {
						if (trueClass == n) {
							// classified incorrectly, but is indeed positive
							// -> false negative
							nFalseNegatives[n]++;
						} else {
							// classified incorrectly, but is indeed negative
							// -> false positive
							nFalsePositives[n]++;
						}
					}
				} // for 0..nClasses
			} // for 0..nInstances
			
			// Show evaluation
			double[] precisionPositiveClass = new double[nClasses];
			double[] precisionNegativeClass = new double[nClasses];
			double[] recallPositiveClass = new double[nClasses];
			double[] recallNegativeClass = new double[nClasses];
			double[] accuracy = new double[nClasses];
			double[] fx = new double[nClasses];
			for (int n = 0; n < nClasses; n++) {
				precisionPositiveClass[n] = nTruePositives[n] / (nTruePositives[n] + nFalsePositives[n]);
				precisionNegativeClass[n] = nTrueNegatives[n] / (nTrueNegatives[n] + nFalseNegatives[n]);
				recallPositiveClass[n] = nTruePositives[n] / (nTruePositives[n] + nFalseNegatives[n]);
				recallNegativeClass[n] = nTrueNegatives[n] / (nTrueNegatives[n] + nFalsePositives[n]);
				accuracy[n] = (nTruePositives[n] + nTrueNegatives[n]) /
					(nTruePositives[n] + nTrueNegatives[n] + nFalsePositives[n] + nFalseNegatives[n]);
				fx[n] = (2 * precisionNegativeClass[n] * recallPositiveClass[n]) /
					(precisionNegativeClass[n] + recallPositiveClass[n]);
				if (talkative) {
					System.out.println("IR stats for biased classifier " + n + ":");
					System.out.println("TP: " + nTruePositives[n] + " FP: " + nFalsePositives[n] +
						" TN: " + nTrueNegatives[n] + " FN: " + nFalseNegatives[n]);
					System.out.println("Positive Class Precision: " + precisionPositiveClass[n] +
						" Recall: " + recallPositiveClass[n]);
					System.out.println("Negative Class Precision: " + precisionNegativeClass[n] +
						" Recall: " + recallNegativeClass[n]);
					System.out.println("Fx: " + fx[n]);
					System.out.println("Training Set Accuracy: " + accuracy[n]);
				}
			}
//			weightAdjustmentDone = true;
			
		}
		
		// TODO self-adjusting weights loop (see above)
		// } while (adjustWeights && wa++ < weightAdjustmentIterations && !weightAdjustmentDone);

		///////////////////////////////////////////////////////////////////
		// 	Train arbiter
		//

		// Arbiter is trained on instances where biased classifiers disagree
		// (or agree, but on the wrong class)
		
		Instances t = new Instances(trainingSet, 0);
		for (int m = 0; m < nInstances; m++) {
			Instance i = trainingSet.instance(m);
			int trueClass = (int) i.classValue();
			double[] details = classifyWithBiasedClassifiers(i);
			Vote v = vote(details);
			
			if (v.agreement < minAgreement || ((v.votedClass != trueClass) && addAgreeWrong)) {

				if (smoCheat) {

					// Inserting one and the same instance multiple times helps SMO
					// (What the heck...?)

					// Instance j = new Instance(i);
					Instance j = (Instance) i.copy();
					j.setWeight(4);
					t.add(j);
				} else if (weightingInsteadOfDropping){
					Instance j = (Instance) i.copy();
					j.setWeight(i.weight() * weightFactorArbiter);
					t.add(j);
				} else {
					t.add(i);
				}
			} else if (weightingInsteadOfDropping) {
				Instance j = (Instance) i.copy();
				j.setWeight(i.weight() / weightFactorArbiter);
				t.add(j);
			}
		}
		
		///////////////////////////////////////////////////////////////////
		// Support for multiple rounds
		//

		if (rounds > 1) {
			if (t.numInstances() == arbiterInstancesInLastRound) {
				if (talkative) {
					System.out.println("Number of instances in dispute set same as in last round," +
						" suspending further rounds.");
				}
				arbiter = Classifier.makeCopies(classifier, 1);
			} else if (t.numInstances() <= 3) {	// hardcoded constant!
				if (talkative && rounds > 1) {
					System.out.println("<= 3 instances in dispute set, suspending further rounds.");
				}
				arbiter = Classifier.makeCopies(classifier, 1);
			} else {
				// everything fine, use another round of Triskel for arbiter				
				arbiter = Classifier.makeCopies(this, 1);
				((TriskelA) arbiter[0]).setRounds(rounds - 1);
				((TriskelA) arbiter[0]).xsetArbiterInstancesInLastRound(t.numInstances());
			}
		} else {
			arbiter = Classifier.makeCopies(classifier, 1);
		}

		if (t.numInstances() <=3) {	// hardcoded constant!
			// If the biased classifiers agreed on every instance (!),
			// then train the arbiter on the complete training set.
			// (What else could we do!?)
			if (talkative) {
				System.out.println("Round " + rounds +
					": Too few instances in dispute set, building arbiter on full training set.");
			} 
			arbiter[0].buildClassifier(trainingSet);
		} else {
			// build arbiter normally
			if (talkative) {
				System.out.println("Round " + rounds +
					": Building arbiter on " + t.numInstances() + " instances.");
			}
			arbiter[0].buildClassifier(t);
		}
	}

	/* (non-Javadoc)
	 * @see weka.core.OptionHandler#listOptions()
	 */
	public Enumeration listOptions() {
		Vector options = new Vector();
		options.add(
				new Option("\tSpecify the full class name of a base classifier " +
						"(default NaiveBayes).", "W", 1, "-W <classname>"));
		options.add(
				new Option("\tSets the method to use. Valid values are " +
						"1 (random undersampling), 2 (overweighting), " +
						"3 (training error) or 4 (direct bias) (default 1)", "M", 1, "-M <int>"));
		options.add(
				new Option("\tRandom number seed when using undersampling",
						"S", 1, "-Q <int>"));
		options.add(
				new Option("\tInstance weight when using oversampling",
						"H", 1, "-H <double>"));
		options.add(
				new Option("\tToggle debug output (0 off or 1 on, default 0)",
						"D", 1, "-D <int>"));
		options.add(
				new Option("\tWhether or not to add instances where the biased classifiers " +
					"agreed on the wrong class to the arbiter (0 off or 1 on, default 0)",
						"Q", 1, "-A <int>"));
		options.add(
				new Option("\tProbability for random undersampling",
						"P", 1, "-P <double>"));
		options.add(
				new Option("\tAutomatically set weights based on the log of the " +
						"number of instances in each class (default 0)",
						"L", 1, "-L <int>"));
		options.add(
				new Option("\tUse 1-vs-All mode for multiclass classification (default 0)",
						"C", 1, "-C <int>"));
		options.add(
				new Option("\tMinimum agreement of classifiers for not asking arbiter (default 1.0)",
						"A", 1, "-A <double>"));
		options.add(
				new Option("\tUse \"SMO Cheat\": Training instances for arbiter are overweighted (default 0)",
						"X", 1, "-X <int>"));
		options.add(
				new Option("\tSpecify the number of rounds for this algorithm. If > 1, the arbiter " +
						"classifier will be replaced with another Triskel.",
						"R", 1, "-R <int>"));
		options.add(
				new Option("\tWhether or not to use weighting instead of dropping instances when " +
					"training the arbiter. (default 1)",
					"U", 1, "-U <int>"));
		options.add(
				new Option("\tWhen weighting for the arbiter instead of dropping instances, use this " +
					"as a factor for weighting. The current weight is multiplied with this number. (default 0.5)",
					"V", 1, "-V <double>"));
		return options.elements();
	}

	/* (non-Javadoc)
	 * @see weka.core.OptionHandler#setOptions(java.lang.String[])
	 */
	public void setOptions(String[] options) throws Exception {
		
		String classifierNameOption = Utils.getOption('W', options);
		
		String methodOption = Utils.getOption('M', options);
		if (methodOption.length() != 0) {
			method = Integer.parseInt(methodOption);
		}
		
		String seedOption = Utils.getOption('S', options);
		if (seedOption.length() != 0) {
			seed = Integer.parseInt(seedOption);
		}
		
		String weightOption = Utils.getOption('H', options);
		if (weightOption.length() != 0 ) {
			oversamplingWeight = Double.parseDouble(weightOption);
		}
		
		String debugOption = Utils.getOption('D', options);
		if (debugOption.length() != 0) {
			int debugOptionInt = Integer.parseInt(debugOption);
			if (debugOptionInt == 1) talkative = true;
		}
		
		String addAgreeWrongOption = Utils.getOption('Q', options);
		if (addAgreeWrongOption.length() != 0) {
			int addAgreeWrongOptionInt = Integer.parseInt(addAgreeWrongOption);
			if (addAgreeWrongOptionInt == 1) addAgreeWrong = true;
		}

		String probOption = Utils.getOption('P', options);
		if (probOption.length() != 0) {
			undersamplingProbability = Double.parseDouble(probOption);
		}
		
		String multiClassOption = Utils.getOption('C', options);
		if (multiClassOption.length() != 0) {
			int multiClassOptionInt = Integer.parseInt(multiClassOption);
			if (multiClassOptionInt == 1) oneVsAll = true;
		}
		
		String logBasedOption = Utils.getOption('L', options);
		if (logBasedOption.length() != 0) {
			int logBasedOptionInt = Integer.parseInt(logBasedOption);
			if (logBasedOptionInt == 1) logBasedWeighting = true;
		}

		String smoCheatOption = Utils.getOption('X', options);
		if (smoCheatOption.length() != 0) {
			int smoCheatOptionInt = Integer.parseInt(smoCheatOption);
			if (smoCheatOptionInt == 1) smoCheat = true;
		}

		String minAgreeOption = Utils.getOption('A', options);
		if (minAgreeOption.length() != 0) {
			minAgreement = Double.parseDouble(minAgreeOption);
		}

		String roundsOption = Utils.getOption('R', options);
		if (roundsOption.length() != 0) {
			rounds = Integer.parseInt(roundsOption);
		}

		String weightingInsteadOption = Utils.getOption('U', options);
		if (weightingInsteadOption.length() != 0) {
			int weightingInsteadOptionInt = Integer.parseInt(weightingInsteadOption);
			if (weightingInsteadOptionInt == 1) weightingInsteadOfDropping = true;
		}

		String weightFactorOption = Utils.getOption('V', options);
		if (weightFactorOption.length() != 0) {
			weightFactorArbiter = Double.parseDouble(weightFactorOption);
		}

		String[] classifierOptions = Utils.partitionOptions(options);
		if (classifierNameOption.length() != 0) {
			classifier = Classifier.forName(classifierNameOption, classifierOptions);
		}
	}

	/* (non-Javadoc)
	 * @see weka.core.OptionHandler#getOptions()
	 */
	public String[] getOptions() {
	    String [] classifierOptions = new String [0];
	    if ((classifier != null) &&
		(classifier instanceof OptionHandler)) {
	      classifierOptions = ((OptionHandler) classifier).getOptions();
	    }
		String [] options = new String [classifierOptions.length + 29];
		int c = 0;
		
		options[c++] = "-W"; options[c++] = classifier.getClass().getName();
		options[c++] = "-M"; options[c++] = "" + method;
		options[c++] = "-P"; options[c++] = "" + undersamplingProbability;
		options[c++] = "-S"; options[c++] = "" + seed;
		options[c++] = "-H"; options[c++] = "" + oversamplingWeight;
		options[c++] = "-A"; options[c++] = "" + minAgreement;
		options[c++] = "-R"; options[c++] = "" + rounds;
		if (addAgreeWrong) {
			options[c++] = "-Q"; options[c++] = "1";
		} else {
			options[c++] = "-Q"; options[c++] = "0";
		}
		if (talkative) {
			options[c++] = "-D"; options[c++] = "1";
		} else {
			options[c++] = "-D"; options[c++] = "0";
		}
		if (smoCheat) {
			options[c++] = "-X"; options[c++] = "1";
		} else {
			options[c++] = "-X"; options[c++] = "0";
		}
		if (logBasedWeighting) {
			options[c++] = "-L"; options[c++] = "1";
		} else {
			options[c++] = "-L"; options[c++] = "0";
		}
		if (oneVsAll) {
			options[c++] = "-C"; options[c++] = "1";
		} else {
			options[c++] = "-C"; options[c++] = "0";
		}
		if (weightingInsteadOfDropping) {
			options[c++] = "-U"; options[c++] = "1";
		} else {
			options[c++] = "-U"; options[c++] = "0";
		}
		options[c++] = "-V"; options[c++] = "" + weightFactorArbiter;
		options[c++] = "--";
		
	    System.arraycopy(classifierOptions, 0, options, c, 
			     classifierOptions.length);
	    c += classifierOptions.length;
	    while (c < options.length) {
	      options[c++] = "";
	    }

		return options;
	}
	
	/**
	 * Main method for testing this class.
	 * Note that if you call the classifier this way, the default settings
	 * are SMO and random undersampling (as opposed to NaiveBayes and
	 * overweighting).
	 */
	public static void main(String[] argv) {

	  TriskelA scheme;

	  try {
		scheme = new TriskelA();
		scheme.setClassifier(new SMO());
		scheme.setMethod(
			new SelectedTag(TriskelA.METHOD_UNDERSAMPLING,
			TriskelA.TAGS_METHOD));
		scheme.setRounds(2);
		scheme.setUndersamplingProbability(0.2);
		System.out.println(Evaluation.evaluateModel(scheme, argv));
	  } catch (Exception e) {
		e.printStackTrace();
		System.err.println(e.getMessage());
	  }
	}

	public Classifier getClassifier() {
		return classifier;
	}

	public void setClassifier(Classifier classifier) {
		this.classifier = classifier;
	}

	public boolean isTalkative() {
		return talkative;
	}

	public void setTalkative(boolean talk) {
		this.talkative = talk;
	}

	public SelectedTag getMethod() {
		return new SelectedTag(method, TAGS_METHOD);
	}

	public void setMethod(SelectedTag newMethod) {
		if (newMethod.getTags() == TAGS_METHOD) {
			method = newMethod.getSelectedTag().getID();
		}
	}

	public String seedTipText() {
		return "Used as seed for random number generator if random undersampling is selected.";
	}

	public int getSeed() {
		return seed;
	}

	public void setSeed(int seed) {
		this.seed = seed;
	}

	public double[] getDetailedResults() {
		return detailedResults;
	}

	public String oversamplingWeightTipText() {
		return "Weight that is assigned to positive instances in oversampling mode";
	}

	public double getOversamplingWeight() {
		return oversamplingWeight;
	}

	public void setOversamplingWeight(double oversamplingWeight) {
		this.oversamplingWeight = oversamplingWeight;
	}

	public String undersamplingProbabilityTipText() {
		return "Probability for adding negative instances in undersampling mode";
	}

	public double getUndersamplingProbability() {
		return undersamplingProbability;
	}

	public void setUndersamplingProbability(double undersamplingProbability) {
		this.undersamplingProbability = undersamplingProbability;
	}

	public String globalInfo() {
		return "An ensemble method that trains specialised classifiers on " +
				"subsets of instances so that effectively the feature space is " +
				"split into sections. (Andreas Hess, University College Dublin," +
				" <andreas.hess@ucd.ie>)";
	}

	/**
	 * @return
	 */
	public boolean isLogBasedWeighting() {
		return logBasedWeighting;
	}

	/**
	 * @return
	 */
	public double getMinAgreement() {
		return minAgreement;
	}

	/**
	 * @return
	 */
	public boolean isOneVsAll() {
		return oneVsAll;
	}

	/**
	 * @param b
	 */
	public void setLogBasedWeighting(boolean b) {
		logBasedWeighting = b;
	}

	/**
	 * @param d
	 */
	public void setMinAgreement(double d) {
		minAgreement = d;
	}

	/**
	 * @param b
	 */
	public void setOneVsAll(boolean b) {
		oneVsAll = b;
	}

	/**
	 * @return
	 */
	public boolean isSmoCheat() {
		return smoCheat;
	}

	/**
	 * @param b
	 */
	public void setSmoCheat(boolean b) {
		smoCheat = b;
	}

	/**
	 * @return
	 */
	public int getRounds() {
		return rounds;
	}

	/**
	 * @param i
	 */
	public void setRounds(int i) {
		rounds = i;
	}

	/**
	 * @return
	 */
	public boolean isAddAgreeWrong() {
		return addAgreeWrong;
	}

	/**
	 * @param b
	 */
	public void setAddAgreeWrong(boolean b) {
		addAgreeWrong = b;
	}

	/**
	 * @param i
	 */
	public void xsetArbiterInstancesInLastRound(int i) {
		arbiterInstancesInLastRound = i;
	}
	
	/**
	 * @return
	 */
	public boolean isAdjustWeights() {
		return adjustWeights;
	}

	/**
	 * @param b
	 */
	public void setAdjustWeights(boolean b) {
		adjustWeights = b;
	}

	/**
	 * @return
	 */
	public double getWeightFactorArbiter() {
		return weightFactorArbiter;
	}

	/**
	 * @return
	 */
	public boolean isWeightingInsteadOfDropping() {
		return weightingInsteadOfDropping;
	}

	/**
	 * @param d
	 */
	public void setWeightFactorArbiter(double d) {
		weightFactorArbiter = d;
	}

	/**
	 * @param b
	 */
	public void setWeightingInsteadOfDropping(boolean b) {
		weightingInsteadOfDropping = b;
	}

}
