/*
 * Created on Dec 2, 2004
 */
package weka.classifiers.ucd.eval;

import java.io.Serializable;
import java.util.Random;

import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.classifiers.ucd.biased.UndersamplingBiasedClassifier;
import weka.core.*;
import weka.core.FastVector;
import weka.core.Instances;
import weka.core.Randomizable;
import weka.core.Tag;

/**
 * This class can be used as an upper class for all meta-classifiers that
 * do some sort of (cross-) evaluation during their training phase.
 * 
 * It provides methods for evaluation according to some user-specified mode.
 * 
 * @author andreas
 */
public abstract class AbstractEvaluatingClassifier
extends Classifier 
implements Randomizable, Serializable {

    /**
     * Random number seed for cross-validation
     */
    protected int seed = 1;
    
    /**
     * Relabeled instances (filled from getPredictions)
     */
    protected Instances relabeledInstances;

    /**
     * Base classifier
     */
    protected Classifier classifier = new UndersamplingBiasedClassifier();
    
    public static final int EVAL_TRAINING_SET = 2;
    public static final int EVAL_TUNED_SPLIT = 1;
    public static final int EVAL_CROSS_VALIDATION = 0;
    public static final Tag [] TAGS_EVAL = {
          new Tag(EVAL_TRAINING_SET, "Evaluate on training set"),
          new Tag(EVAL_TUNED_SPLIT, "Evaluate on single fold"),
          new Tag(EVAL_CROSS_VALIDATION, "N-Fold cross validation")
        };
    
    /**
     * The mode of evaluation
     */
    protected int evaluationMode = EVAL_CROSS_VALIDATION;
    
    /**
     * Number of folds for cross-validation
     */
    protected int numFolds = 3;

    /**
     * In checkForInstance it is checked (when doing evalutation on a single
     * fold) if both the training and the testing split contain at least one
     * instance of this class.
     */
    protected int classIndex = 0;
    
    /**
     * Maximum number of rounds for optimization.
     * Should be evaluated in the buildClassifier method of implementing
     * classes!!
     */
    protected int maxRounds = 3;
    
    /**
     * A random number generator. Should be initialized in buildClassifier!!
     */
    protected Random rnd;

    /**
     * Gets predictions and optionally relabels misclassified instances of the
     * preferred class as instances of the opposite class 
     *
     * @param instances the set of <code>Instances</code> to generate
     * predictions for.
     * @return a <code>FastVector</code> containing the predictions.
     * @exception Exception if an error occurs generating the predictions.
     */
    protected FastVector getPredictions(Instances instances, boolean relabel) 
    throws Exception {
        RelabelUtils eu;
        if (relabel) {
            eu = new RelabelUtils(instances, classIndex, (classIndex + 1) % 2);
        } else {
            eu = new RelabelUtils();
        }
        eu.setSeed(seed);
        return getPredictions(instances, eu);
    }
    
    protected FastVector getPredictions(Instances instances) 
    throws Exception {
        return getPredictions(instances, false);
    }
    
    protected FastVector getPredictions(Instances instances, RelabelUtils eu) 
    throws Exception {
        switch (evaluationMode) {
        case EVAL_TUNED_SPLIT:
            Instances trainData = null, evalData = null;
            Instances data = new Instances(instances);
            data.randomize(rnd);
            data.stratify(numFolds);
            
            // Make sure that both subsets contain at least one positive instance
            for (int subsetIndex = 0; subsetIndex < numFolds; subsetIndex++) {
                trainData = data.trainCV(numFolds, subsetIndex, rnd);
                evalData = data.testCV(numFolds, subsetIndex);
                if (checkForInstance(trainData) && checkForInstance(evalData)) {
                    break;
                }
            }
            FastVector r = eu.getTrainTestPredictions(classifier, trainData, evalData);
            relabeledInstances = eu.getRelabeledInstances();
            return r;
        case EVAL_TRAINING_SET:
            r = eu.getTrainTestPredictions(classifier, instances, instances);
            relabeledInstances = eu.getRelabeledInstances();
            return r;
        case EVAL_CROSS_VALIDATION:
            r = eu.getCVPredictions(classifier, instances, numFolds);
            relabeledInstances = eu.getRelabeledInstances();
            return r;
        default:
            throw new RuntimeException("Unrecognized evaluation mode");
        }
    }

    /**
     * Evaluate a classifier with the current settings.
     * This method is inspired by getPredictions(), but returns an Evaluation
     * object rather than Predictions.
     * 
     * @param instances The instance set to operate on
     * @return An Evaluation object with the desired evaluation performed
     * @throws Exception If the evaluation mode is unrecognized or something else goes wrong during classification.
     */
    protected Evaluation getEvaluation(Instances instances) throws Exception {
        
        Evaluation eu = new Evaluation(instances);
        
        switch (evaluationMode) {
        case EVAL_TUNED_SPLIT:
            Instances trainData = null, evalData = null;
            Instances data = new Instances(instances);
            data.randomize(rnd);
            data.stratify(numFolds);
            
            // Make sure that both subsets contain at least one instance
            // of each class
            for (int subsetIndex = 0; subsetIndex < numFolds; subsetIndex++) {
                trainData = data.trainCV(numFolds, subsetIndex, rnd);
                evalData = data.testCV(numFolds, subsetIndex);
                if (checkForInstance(trainData) && checkForInstance(evalData)) {
                    break;
                }
            }
            classifier.buildClassifier(trainData);
            eu.evaluateModel(classifier, evalData);
            return eu;
        case EVAL_TRAINING_SET:
            classifier.buildClassifier(instances);
            eu.evaluateModel(classifier, instances);
            return eu;
        case EVAL_CROSS_VALIDATION:
            eu.crossValidateModel(classifier, instances, numFolds, rnd);
            return eu;
        default:
            throw new RuntimeException("Unrecognized evaluation mode");
        }
    }

    /**
     * Checks whether instance of designated class is in subset.
     */
    private boolean checkForInstance(Instances data) throws Exception {
    
        for (int i = 0; i < data.numInstances(); i++) {
            if (((int)data.instance(i).classValue()) == classIndex) {
                return true;
            }
        }
        return false;
    }

    public void setSeed(int seed) {
        this.seed = seed;
    }

    public int getSeed() {
        return seed;
    }

    /**
     * @return Returns the maxRounds.
     */
    public int getMaxRounds() {
        return maxRounds;
    }

    /**
     * @param maxRounds The maxRounds to set.
     */
    public void setMaxRounds(int maxRounds) {
        this.maxRounds = maxRounds;
    }

    /**
     * @return Returns the numFolds.
     */
    public int getNumFolds() {
        return numFolds;
    }

    /**
     * @param numFolds The numFolds to set.
     */
    public void setNumFolds(int numFolds) {
        this.numFolds = numFolds;
    }

    /**
     * Sets the evaluation mode used. Will be one of
     * EVAL_TRAINING, EVAL_TUNED_SPLIT, or EVAL_CROSS_VALIDATION
     *
     * @param newMethod the new evaluation mode.
     */
    public void setEvaluationMode(SelectedTag newMethod) {
      
      if (newMethod.getTags() == TAGS_EVAL) {
        evaluationMode = newMethod.getSelectedTag().getID();
      }
    }

    /**
     * Gets the evaluation mode used. Will be one of
     * EVAL_TRAINING, EVAL_TUNED_SPLIT, or EVAL_CROSS_VALIDATION
     *
     * @return the evaluation mode.
     */
    public SelectedTag getEvaluationMode() {

      return new SelectedTag(evaluationMode, TAGS_EVAL);
    }

}
