package weka.classifiers.ucd.eval;

import java.util.Random;

import weka.classifiers.Classifier;
import weka.classifiers.evaluation.NumericPrediction;
import weka.classifiers.evaluation.Prediction;
import weka.core.*;

/**
 * Class that is almost the same as EvaluationUtils, but, while it performs
 * classification, it creates a new instance set where misclassified instances
 * of a specified class (the "preferred" class) are re-labeled.
 * @author andreas
 */
public class RelabelUtils implements Randomizable {

    /**
     * Random number seed for cross-validation
     */
    protected int seed = 1;
    
    /**
     * The preferred class index
     */
    // protected int classIndex = 0;
    
    /**
     * Misclassified instances that are actually of the preferred class
     * are re-labeled with this class index.
     * If the class index is -1, then the instances are not inserted in
     * the set of relabeled instances.
     */
    // protected int relabelAs = 1;
    
    protected double[][] relabelMatrix;
    
    /**
     * The re-labeled instances are stored in here
     */
    protected Instances relabeledInstances;

    /**
     * Create a new RelabelUtils object 
     * @param instances The instances that we're working on here
     * @param _preferredClass The preferred class
     * @param _relabelAs The preferred class, if misclassified, is relabeled with this class index.
     */
    public RelabelUtils(Instances instances,
            int _preferredClass, int _relabelAs) { 
        relabeledInstances = new Instances(instances, 0);
    	int nClasses = instances.numClasses();
    	//                      predicted v  v actual
    	relabelMatrix = new double[nClasses][nClasses];
    	for (int n = 0; n < nClasses; n++) {
    		for (int m = 0; m < nClasses; m++) {
    			if (n == _preferredClass && m != _preferredClass) {
    				relabelMatrix[n][m] = _relabelAs;
    			} else {
    				relabelMatrix[n][m] = n;
    			}
    		}
    	}
    }
    
    public RelabelUtils(Instances _relabeledInstances, double[][] _matrix) {
    	relabelMatrix = _matrix;
    	relabeledInstances = _relabeledInstances;
    }
    
    public RelabelUtils() {
    	relabelMatrix = null;
    	relabeledInstances = null;
    }
    
    /* (non-Javadoc)
     * @see weka.core.Randomizable#setSeed(int)
     */
    public void setSeed(int seed) {
        this.seed = seed;
    }

    /* (non-Javadoc)
     * @see weka.core.Randomizable#getSeed()
     */
    public int getSeed() {
        return seed;
    }
    
    /**
     * Generate a bunch of predictions ready for processing, by performing a
     * cross-validation on the supplied dataset.
     *
     * @param classifier the Classifier to evaluate
     * @param data the dataset
     * @param numFolds the number of folds in the cross-validation.
     * @exception Exception if an error occurs
     */
    public FastVector getCVPredictions(Classifier classifier, 
                                       Instances data, 
                                       int numFolds) 
      throws Exception {

      FastVector predictions = new FastVector();
      Instances runInstances = new Instances(data);
      Random random = new Random(seed);
      runInstances.randomize(random);
      if (runInstances.classAttribute().isNominal() && (numFolds > 1)) {
        runInstances.stratify(numFolds);
      }
      int inst = 0;
      for (int fold = 0; fold < numFolds; fold++) {
        Instances train = runInstances.trainCV(numFolds, fold, random);
        Instances test = runInstances.testCV(numFolds, fold);
        FastVector foldPred = getTrainTestPredictions(classifier, train, test);
        predictions.appendElements(foldPred);
      } 
      return predictions;
    }

    /**
     * Generate a bunch of predictions ready for processing, by performing a
     * evaluation on a test set after training on the given training set.
     *
     * @param classifier the Classifier to evaluate
     * @param train the training dataset
     * @param test the test dataset
     * @exception Exception if an error occurs
     */
    public FastVector getTrainTestPredictions(Classifier classifier, 
                                              Instances train, Instances test) 
      throws Exception {
      
      classifier.buildClassifier(train);
      return getTestPredictions(classifier, test);
    }

    /**
     * Generate a bunch of predictions ready for processing, by performing a
     * evaluation on a test set assuming the classifier is already trained.
     *
     * @param classifier the pre-trained Classifier to evaluate
     * @param test the test dataset
     * @exception Exception if an error occurs
     */
    public FastVector getTestPredictions(Classifier classifier, 
                                         Instances test) 
      throws Exception {
      
      FastVector predictions = new FastVector();
      for (int i = 0; i < test.numInstances(); i++) {
        if (!test.instance(i).classIsMissing()) {
          predictions.addElement(getPrediction(classifier, test.instance(i)));
        }
      }
      return predictions;
    }
   
    /**
     * Generate a single prediction for a test instance given the pre-trained
     * classifier.
     *
     * @param classifier the pre-trained Classifier to evaluate
     * @param test the test instance
     * @exception Exception if an error occurs
     */
    public Prediction getPrediction(Classifier classifier,
                                    Instance test)
      throws Exception {
     
    	double actual = test.classValue();
    	double [] dist = classifier.distributionForInstance(test);
    	if (test.classAttribute().isNominal()) {
    		NominalInstancePrediction np = new NominalInstancePrediction(test, actual, dist, test.weight());
    		if (relabeledInstances != null) {
	    		Instance i = (Instance) test.copy();
	    		if (relabelMatrix != null) {
	    			double newLabel = relabelMatrix[(int) np.predicted()][(int) actual];
	    			if (newLabel != -1) {
		    			i.setClassValue(newLabel);
			    		relabeledInstances.add(i);
	    			}
	    		} else {
		    		relabeledInstances.add(i);
	    			
	    		}
    		}
    		return np;
    	} else {
    		return new NumericPrediction(actual, dist[0], test.weight());
    	}
    }
    
    /**
     * @return Returns the relabeledInstances.
     */
    public Instances getRelabeledInstances() {
        return relabeledInstances;
    }
    
    /**
     * @param relabeledInstances The relabeledInstances to set.
     */
    public void setRelabeledInstances(Instances relabeledInstances) {
        this.relabeledInstances = relabeledInstances;
    }
}
